from copy import copy
from typing import Tuple, Union

import numpy as np
from pytest import mark

from pyopencl_extension import Helpers, Kernel, Program, Global, Scalar, Types, Array, empty, zeros, to_device

__author__ = "P.Mohr"
__copyright__ = "26.03.2020, P.Mohr"
__version__ = "1.0"
__email__ = "philipp.mohr@tuhh.de"
__doc__ = """This script implements the functionality to copy certain region_in of cl array on device."""

from pyopencl_extension.types.utilities_np_cl import c_name_from_dtype

TypeSliceFormatCopyArrayRegion = Tuple[Tuple[int, Union[int, None], Union[int, None], int], ...]


# supports sclicing with [start,stop] and [start:stop:step_width]


class _Slice:
    def __getitem__(self, val) -> TypeSliceFormatCopyArrayRegion:
        def convert(axis, _val):
            return (axis,
                    0 if _val.start is None else _val.start,
                    _val.stop,
                    1 if _val.step is None else _val.step)

        if not isinstance(val, tuple):
            return convert(0, val),
        else:
            return tuple([convert(axis, _val) for axis, _val in enumerate(val)])


Slice = _Slice()


def test_slice():
    res = Slice[2:]
    assert res == ((0, 2, None, 1),)
    res = Slice[2::3]
    assert res == ((0, 2, None, 3),)
    res = Slice[2:10:3]
    assert res == ((0, 2, 10, 3),)
    res = Slice[2:, 3:-1:2, :]
    assert res == ((0, 2, None, 1),
                   (1, 3, -1, 2),
                   (2, 0, None, 1))
    res = Slice[0:0]
    assert res == ((0, 0, 0, 1),)
    res = Slice[:0]
    assert res == ((0, 0, 0, 1),)
    res = Slice[:, :0]
    assert res == ((0, 0, 0, 1),)


class CopyArrayRegion:
    def _command_for_addr_out_computation(self):
        command = '0'
        for i in range(self.in_buffer.ndim):
            offset = '1'
            for j in range(i + 1, self.in_buffer.ndim):
                offset += '*{}'.format(self.out_buffer.shape[j])
            command += '+(get_global_id({})+{})*{}'.format(i, self.region_out[i][0], offset)
        return command

    def _command_for_addr_in_computation(self):
        command = '0'
        for i in range(self.in_buffer.ndim):
            offset = '1'
            for j in range(i + 1, self.in_buffer.ndim):
                offset += '*{}'.format(self.in_buffer.shape[j])
            command += '+(get_global_id({})+{})*{}'.format(i, self.region_in[i][0], offset)
        return command

    @staticmethod
    def _deal_with_incomplete_regions(region, array: Array):
        # deal with incomplete regions. Fill remaining ranges for axes from input buffer size
        reg = [(0, ax_dim, 1) for ax_dim in array.shape]
        for ax in region:
            reg[ax[0]] = ax[1:]
        return reg

    @staticmethod
    def _deal_with_none_in_stop(region, array: Array):
        # deal with None in stop e.g. Slice[:] -> ((0, 0, None, 1),)
        return [[ax[0], ax[1] if ax[1] is not None else array.shape[i], ax[2]]
                for i, ax in enumerate(region)]

    @staticmethod
    def _deal_with_negative_region(region, array: Array):
        # deal with negative regions
        return [[idx if idx >= 0 else array.shape[i] + idx for idx in ax] for i, ax in
                enumerate(region)]

    def __init__(self, in_buffer: Array,
                 region_in: TypeSliceFormatCopyArrayRegion = None,
                 out_buffer: Array = None,
                 region_out: TypeSliceFormatCopyArrayRegion = None):
        """

        :param in_buffer:
        :param region_in: e.g. region_in=((0,0,1,2),(1,1,3,2)) selects the region_in of array a, like numpy would do with
        a[0:1:2,1:3:2] where 2 is the step width. The first element of tuple selects the axis.
        :param out_buffer: target buffer where data from in_buffer is being copied. (optional)
        :param region_out: specifies the region of out buffer memory where in_buffer data is copied (optional)
        """
        _region_in_original = copy(region_in)  # for debug purposes
        _region_out_original = copy(region_out)  # for debug purposes

        if region_in is not None:
            region_in = self._deal_with_incomplete_regions(region_in, in_buffer)

        if out_buffer is not None and region_out is not None:
            region_out = self._deal_with_incomplete_regions(region_out, out_buffer)

        if region_in is not None:
            region_in = self._deal_with_none_in_stop(region_in, in_buffer)
        if region_out is not None and out_buffer is not None:
            region_out = self._deal_with_none_in_stop(region_out, out_buffer)

        if region_in is not None:
            region_in = self._deal_with_negative_region(region_in, in_buffer)
        if region_out is not None and out_buffer is not None:
            region_out = self._deal_with_negative_region(region_out, out_buffer)

        self.in_buffer = in_buffer
        if region_in is None:
            self.region_in = [(0, self.in_buffer.shape[i_axis], 1) for i_axis in range(self.in_buffer.ndim)]
        else:
            self.region_in = region_in

        if out_buffer is None and region_out is None:
            shape = [ax[1] - ax[0] for ax in self.region_in]
            self.out_buffer = empty(tuple(shape), dtype=self.in_buffer.dtype)
            self.region_out = [(0, i, 1) for i in shape]  # (tuple([0]*len(shape)),shape)
        elif out_buffer is not None and region_out is not None:
            self.out_buffer = out_buffer
            self.region_out = region_out
        else:
            raise ValueError('Case of input argument combination not supported')

        if self.in_buffer.dtype != self.out_buffer.dtype:
            raise ValueError('in and out buffer must be of same type')

        self.shape_region_out = tuple([ax[1] - ax[0] for ax in self.region_out])
        self.in_buffer = in_buffer

        self.copy_array_region = Kernel(name='copy_array_region',
                                        args={'in_buffer': Global(self.in_buffer, read_only=True),
                                              'out_buffer': Global(self.out_buffer, )},
                                        body=["""
                                  int addr_in = ${command_addr_in};
                                  int addr_out = ${command_addr_out};
                                  out_buffer[addr_out]=in_buffer[addr_in];
                                  """],
                                        replacements={'command_addr_in': self._command_for_addr_in_computation(),
                                                      'command_addr_out': self._command_for_addr_out_computation()},
                                        global_size=self.shape_region_out).compile()

    def __call__(self, *args, **kwargs):
        self.copy_array_region()
        return self.out_buffer


def cl_set(array: Array, region: TypeSliceFormatCopyArrayRegion, value):
    """
    example usage:
    set slice of array with scalar value
    val = 1
    cl_set(ary, Slice[:,2:3], val)

    set slice of array with equally shaped numpy array like the slice
    some_np_array = np.array([[3,4])
    cl_set(ary, Slice[1:2,2:3], some_np_array)


    :param array:
    :param region:
    :param value:
    :return:
    """
    # todo test if array c contiguous
    region_arg = region
    # if slice is contiguous block of memory set it as
    # _buffer_np = np.zeros_like(add_symbols_memory_initialization.out_buffer)
    # _buffer_np[:, memory: -memory] = mapper.alphabet[0]
    # add_symbols_memory_initialization.out_buffer.set(_buffer_np)
    region = CopyArrayRegion._deal_with_incomplete_regions(region_arg, array)
    region = CopyArrayRegion._deal_with_none_in_stop(region, array)
    region = CopyArrayRegion._deal_with_negative_region(region, array)

    # test if requested region is
    for axis, _slice in enumerate(region):
        step_width = _slice[2]
        if abs(_slice[0] * step_width) > array.shape[axis] or abs(_slice[1] * step_width) > array.shape[axis]:
            raise ValueError('Slicing out of array bounds')
    if any([(part[0] - part[1]) == 0 for part in region]):  # check that there is no empty slice
        return
    global_size = np.product([part[1] - part[0] for part in region])
    target_shape = to_device(np.array(array.shape).astype(Types.int))
    offset_target = to_device(np.array([part[0] for part in region]).astype(Types.int))
    source_shape = to_device(np.array([part[1] - part[0] for part in region]).astype(Types.int))
    source_n_dims = len(source_shape)

    if isinstance(value, np.ndarray):
        source = to_device(value.astype(array.dtype))
        arg_source = Global(source)
        code_source = 'source[get_global_id(0)]'
    else:
        arg_source = Scalar(array.dtype)
        source = value
        code_source = 'source'

    knl = Kernel('set_cl_array',
                 {'target': Global(array),
                  'target_shape': Global(target_shape),
                  'offset_target': Global(offset_target),
                  'source': arg_source,
                  'source_shape': Global(source_shape),
                  'source_n_dims': Scalar(Types.int)},
                 """
   // id_source = get_global_id(0)
   // id_source points to element of array source which replaces element with id_target in array target.
   // we need to compute id_target from id_source:
   // we assume c-contiguous addressing like:
   // id_source = id0*s1*s2*s3+id1*s2*s3+id2*s3+id3 (here s refers shape of source array)
   // At first we need to compute individual ids of source array from id_source:
   // id3 = int(gid % s3), temp = (gid-id3)/s3
   // id2 = int(temp % s2), temp = (temp-id2)/s2
   // id1 = int(temp % s1), temp = (temp-id1)/s1
   // id0 = int(temp % s0), temp = (temp-id0)/s1
   // Finally, we can determine the id of the target array and copy element to corresponding position:
   // id_target = (id0*offset0t)*s1t*s2t ... (sxt: shape of target array along dim x)
   int id_target = 0; // to be iteratively computed from global id, slice dimensions and ary dimensions

   int temp = get_global_id(0);
   int prod_source_id_multiplier = 1;
   int prod_target_id_multiplier = 1;

   for(int i=source_n_dims-1; i>=0; i--){ // i=i_axis_source
    int id_source = temp % source_shape[i];
    temp = (int)((temp-id_source)/source_shape[i]);
    prod_source_id_multiplier *= source_shape[i];
    id_target += (offset_target[i]+id_source)*prod_target_id_multiplier;
    prod_target_id_multiplier *= target_shape[i];
   }
   target[id_target] = ${source};
                   """,
                 replacements={'addr': Helpers.command_compute_address(array.ndim),
                               'source': code_source},
                 global_size=(global_size,)
                 ).compile(array.context, emulate=False)
    knl(source=source, source_n_dims=source_n_dims)


@mark.parametrize('s', [[0, 1, 2, 3],
                        [0, 10, 1, 3],
                        [4, 8, 0, 3],
                        [4, -1, 0, 3],
                        [0, 10, 0, 3]])
def test_cl_set(s):
    ary = zeros((10, 3), Types.double)
    ary_np = ary.get()
    val = 2
    # ClSet(ary)[:, 2:3] = val
    ary_np[s[0]:s[1], s[2]:s[3]] = val
    cl_set(ary, Slice[s[0]:s[1], s[2]:s[3]], val)

    assert np.all(ary_np == ary.get())


def test_cl_set_out_of_bounds():
    ary = zeros((10, 3), Types.double)
    val = 2
    try:
        cl_set(ary, Slice[:100, :], val)
    except ValueError as err:
        assert str(err) == 'Slicing out of array bounds'


def test_cl_set_many_dimensions():
    s = [0, 1, 2, 3, 0, 1]
    ary = zeros((10, 3, 4), Types.double)
    ary_np = ary.get()
    val = 2
    # ClSet(ary)[:, 2:3] = val
    ary_np[s[0]:s[1], :, s[4]:s[5]] = val
    cl_set(ary, Slice[s[0]:s[1], :, s[4]:s[5]], val)
    assert np.all(ary_np == ary.get())


class TypeConverter:
    def __init__(self, in_buffer: Array, out_buffer_dtype: np.dtype):
        self.in_buffer = in_buffer
        self.out_buffer = empty(in_buffer.shape, out_buffer_dtype, in_buffer.queue)
        knl = Kernel(name='type',
                     args={'in_buffer': Global(self.in_buffer, read_only=True),
                           'out_buffer': Global(self.out_buffer)},
                     body=["""
                                   int addr_in = ${command_addr_in};
                                   int addr_out = ${command_addr_out};
                                   out_buffer[addr_out]=convert_${buff_out_t}(in_buffer[addr_in]);
                                   """],
                     replacements={'command_addr_in': Helpers.command_compute_address(self.in_buffer.ndim),
                                   'command_addr_out': Helpers.command_compute_address(self.out_buffer.ndim),
                                   'buff_out_t': c_name_from_dtype(self.out_buffer.dtype)},
                     global_size=self.in_buffer.shape)
        self.program = Program(kernels=[knl]).compile(context=in_buffer.context, emulate=False)

    def __call__(self):
        self.program.type()
        # self.in_buffer.queue.finish()
        return self.out_buffer



import logging
import os
import time
from dataclasses import dataclass
from enum import Enum

import numpy as np
import pyopencl as cl

from pyopencl_extension import get_devices, Context


class QueueProperties(Enum):
    DEFAULT = 0
    ON_DEVICE = 4
    ON_DEVICE_DEFAULT = 8
    OUT_OF_ORDER_EXEC_MODE_ENABLE = 1
    # https://stackoverflow.com/questions/29068229/is-there-a-way-to-profile-an-opencl-or-a-pyopencl-program
    PROFILING_ENABLE = 2


@dataclass
class TimingsQueue:
    queue_created: int = 0
    compilation: int = 0
    blocking: int = 0
    profiling_finished: int = 0
    first_event_added: int = 0
    last_event_added: int = 0
    profiling_enabled: bool = True

    def __post_init__(self):
        self.queue_created = time.perf_counter_ns()

    def add_blocking(self, t):
        if self.profiling_enabled:
            self.blocking += t


class CommandQueue(cl.CommandQueue):
    def __str__(self):
        return f'Queue: {self.context.device_id=}'

    def __init__(self, context, device=None, properties: QueueProperties | int = QueueProperties.DEFAULT,
                 max_len_events=1e6):
        if isinstance(properties, QueueProperties):
            properties = properties.value
        super().__init__(context, device, properties)
        self.max_len_events = int(max_len_events)
        self.events = []
        self._context = context  # inside of super call context is casted to original pyopencl context class
        self.profiling_enabled = True if self.properties == cl.command_queue_properties.PROFILING_ENABLE else False
        self.t_ns = TimingsQueue(profiling_enabled=self.profiling_enabled)

    @property
    def context(self) -> Context:
        return self._context

    def get_profiler(self) -> 'Profiling':
        self.t_ns.profiling_finished = time.perf_counter_ns()
        self.t_ns.compilation = self.context.time_compilation_ns
        return Profiling(self)

    def add_event(self, event, name):
        if self.profiling_enabled:
            if len(self.events) == 0:
                self.t_ns.first_event_added = time.perf_counter_ns()
            if len(self.events) < self.max_len_events:
                self.events.append((name, event))
                self.t_ns.last_event_added = time.perf_counter_ns()
            else:
                raise ValueError('Forgot to disable profiling?')

    def finish(self):
        t0 = time.perf_counter_ns()
        super().finish()
        self.t_ns.add_blocking(time.perf_counter_ns() - t0)


class Profiling:
    def __init__(self, queue: CommandQueue):
        if queue.properties == cl.command_queue_properties.PROFILING_ENABLE:
            self.queue = queue
            self.events = queue.events
        else:
            raise ValueError('Profiling must be enabled using command_queue_properties.PROFILING_ENABLE')

    def get_event_names_time_ms(self):
        return [(event[0], 1e-6 * ((_ := event[1]).profile.end - _.profile.start)) for event in self.events]

    def list_cumulative_event_times_ms(self):
        knl_names = {_[0]: 0.0 for _ in self.get_event_names_time_ms()}
        for _ in self.get_event_names_time_ms():
            knl_names[_[0]] += _[1]
        indices = np.argsort([knl_names[_] for _ in knl_names])
        return [list(knl_names.items())[i] for i in np.flip(indices)]

    def get_sum_execution_time(self):
        return sum([_[1] for _ in self.list_cumulative_event_times_ms()])

    def show_histogram_cumulative_kernel_times(self):
        self.queue.finish()
        import matplotlib.pyplot as plt
        import numpy as np
        fig, ax = plt.subplots()
        fig.subplots_adjust(left=0.4)
        # Example data

        event_names = [_[0] for _ in self.list_cumulative_event_times_ms()]
        event_times = [_[1] for _ in self.list_cumulative_event_times_ms()]
        event_names_times = [item for item in zip(event_names, event_times)]
        t_ns = self.queue.t_ns
        t0 = self.events[0][1].profile.start - (t_ns.first_event_added - t_ns.queue_created)
        t1 = (_ := self.events[-1][1].profile).end + (t_ns.profiling_finished - t_ns.last_event_added -
                                                      (_.end - _.start))

        total_time_ms = (t_ns.profiling_finished - t_ns.queue_created) / 1e6
        time_blocking_ms = t_ns.blocking / 1e6  # time waiting for kernels/transfers to finish
        time_gross_python_ms = total_time_ms - time_blocking_ms

        starts = np.array([e[1].profile.start for e in self.events] + [t1])
        ends = np.array([t0] + [e[1].profile.end for e in self.events])
        time_python_net_ms = (starts - ends).sum() / 1e6 - time_blocking_ms

        event_names_times.append(('total opencl events', self.get_sum_execution_time()))
        event_names_times.append(('total overall program', total_time_ms))
        event_names_times.append(('opencl blocking', time_blocking_ms))
        # event_names_times.append(('python (gross)', time_gross_python_ms))
        event_names_times.append(('python (net)', time_python_net_ms))
        event_names_times.append(('compilation', t_ns.compilation / 1e6))

        y_pos = np.arange(len(event_names_times))
        ax.barh(y_pos, [i[1] for i in event_names_times], align='center')
        ax.set_yticks(y_pos)
        ax.set_yticklabels([i[0] for i in event_names_times])
        ax.invert_yaxis()  # labels read top-to-bottom
        ax.set_xlabel('Time ms')
        plt.show()

    # todo: make event plot with rectangles corresponding to kernel times
    # def show_events(self):
    #     import numpy as np
    #     t0 = self.events[0][1].profile.start
    #     starts = np.array([[e[1].profile.start - t0 for e in self.events]])
    #     ends = np.array([[e[1].profile.end - t0 for e in self.events]])
    #     durations_ms = (ends-starts)/1e6
    #
    #     import matplotlib.pyplot as plt
    #     import matplotlib
    #     matplotlib.rcParams['font.size'] = 8.0
    #     plt.figure()
    #     # create a vertical plot
    #     time_center = starts  # np.array([[0.25, 1.0, 1.5]])
    #     time_length = durations_ms / 10  # np.array([[0.5, 0.3, 0.4]]) * 300
    #     plt.eventplot(time_center, colors='blue', lineoffsets=1,
    #                   linelengths=1, linewidths=time_length, orientation='horizontal')
    #     plt.show()


def get_context(device_id: int = None):
    """

    This function facilitates to get a context and queue pointing to a particular device.
    :return: the context instance
    """
    if device_id is None:
        context = cl.create_some_context()
    else:  # currently only a single device is supported. If required interfac must be adjusted to accept multiple ids
        device = get_devices()[device_id]
        context = Context(devices=[device])
    return context


def get_device_id_from_env_var() -> int:
    # add environmental variable PYOPENCL_DEVICE with 0 to select device 0 as default device
    try:
        device_id = os.environ["PYOPENCL_DEVICE"]
        return int(device_id)
    except KeyError:
        logging.info('Default device not configured in environmental variables.')
        return 0


# Convenience feature to access a global queue instance, e.g. useful to avoid passing queue into functions.
_current_queue = None
# If queue is requested via get_current_queue and _current_queue is None, then the default
# device specified below is used and is set as the current queue.
_default_device = get_device_id_from_env_var()


def set_current_queue(queue: CommandQueue | None):
    global _current_queue
    _current_queue = queue
    logging.debug(f'set_current_queue {queue}')


def set_default_device(device_id: int):
    """
    Calling this function overrides the default device (may override what has been set with env var
    os.environ["PYOPENCL_DEVICE"]).
    """
    global _default_device
    _default_device = device_id
    logging.debug(f'set_default_device {device_id=}')


def create_queue(device_id: int = None, queue_properties: QueueProperties = QueueProperties.DEFAULT,
                 compiler_output: bool = False, context: Context = None, *args, **kwargs) -> CommandQueue:
    """
    A convenience function which creates a queue object.

    :return: A container class with context and queue pointing to selected device.
    """
    if device_id is None:
        device_id = _default_device
    if context is None:
        context = get_context(device_id)
    queue = CommandQueue(context, properties=queue_properties)
    queue.finish()

    os.environ['PYOPENCL_COMPILER_OUTPUT'] = '1' if compiler_output else '0'
    set_current_queue(queue)
    logging.info(f'Create queue with {device_id=}')
    return queue


def get_current_queue(*args, **kwargs) -> CommandQueue:
    global _current_queue
    if _current_queue is None:
        _current_queue = create_queue(_default_device, *args, **kwargs)
    return _current_queue


def get_device(device_id: int) -> cl.Device:
    return get_devices()[device_id]


def activate_profiling(**kwargs):
    set_current_queue(create_queue(queue_properties=QueueProperties.PROFILING_ENABLE, **kwargs))


def evaluate_profiling():
    get_current_queue().get_profiler().show_histogram_cumulative_kernel_times()


from pyopencl_extension.modifications_pyopencl import cltypes
from dataclasses import dataclass
import numpy as np
from typing import Callable, Union

@dataclass(frozen=True)
class ClTypesVector:
   char2:Union[np.dtype, Callable]=cltypes.char2
   char4:Union[np.dtype, Callable]=cltypes.char4
   char8:Union[np.dtype, Callable]=cltypes.char8
   char16:Union[np.dtype, Callable]=cltypes.char16
   short2:Union[np.dtype, Callable]=cltypes.short2
   short4:Union[np.dtype, Callable]=cltypes.short4
   short8:Union[np.dtype, Callable]=cltypes.short8
   short16:Union[np.dtype, Callable]=cltypes.short16
   int2:Union[np.dtype, Callable]=cltypes.int2
   int4:Union[np.dtype, Callable]=cltypes.int4
   int8:Union[np.dtype, Callable]=cltypes.int8
   int16:Union[np.dtype, Callable]=cltypes.int16
   long2:Union[np.dtype, Callable]=cltypes.long2
   long4:Union[np.dtype, Callable]=cltypes.long4
   long8:Union[np.dtype, Callable]=cltypes.long8
   long16:Union[np.dtype, Callable]=cltypes.long16
   uchar2:Union[np.dtype, Callable]=cltypes.uchar2
   uchar4:Union[np.dtype, Callable]=cltypes.uchar4
   uchar8:Union[np.dtype, Callable]=cltypes.uchar8
   uchar16:Union[np.dtype, Callable]=cltypes.uchar16
   ushort2:Union[np.dtype, Callable]=cltypes.ushort2
   ushort4:Union[np.dtype, Callable]=cltypes.ushort4
   ushort8:Union[np.dtype, Callable]=cltypes.ushort8
   ushort16:Union[np.dtype, Callable]=cltypes.ushort16
   uint2:Union[np.dtype, Callable]=cltypes.uint2
   uint4:Union[np.dtype, Callable]=cltypes.uint4
   uint8:Union[np.dtype, Callable]=cltypes.uint8
   uint16:Union[np.dtype, Callable]=cltypes.uint16
   ulong2:Union[np.dtype, Callable]=cltypes.ulong2
   ulong4:Union[np.dtype, Callable]=cltypes.ulong4
   ulong8:Union[np.dtype, Callable]=cltypes.ulong8
   ulong16:Union[np.dtype, Callable]=cltypes.ulong16
   half2:Union[np.dtype, Callable]=cltypes.half2
   half4:Union[np.dtype, Callable]=cltypes.half4
   half8:Union[np.dtype, Callable]=cltypes.half8
   half16:Union[np.dtype, Callable]=cltypes.half16
   float2:Union[np.dtype, Callable]=cltypes.float2
   float4:Union[np.dtype, Callable]=cltypes.float4
   float8:Union[np.dtype, Callable]=cltypes.float8
   float16:Union[np.dtype, Callable]=cltypes.float16
   double2:Union[np.dtype, Callable]=cltypes.double2
   double4:Union[np.dtype, Callable]=cltypes.double4
   double8:Union[np.dtype, Callable]=cltypes.double8
   double16:Union[np.dtype, Callable]=cltypes.double16

@dataclass(frozen=True)
class ClTypesScalar:
   char:Union[np.dtype, Callable]=cltypes.char
   short:Union[np.dtype, Callable]=cltypes.short
   int:Union[np.dtype, Callable]=cltypes.int
   long:Union[np.dtype, Callable]=cltypes.long
   uchar:Union[np.dtype, Callable]=cltypes.uchar
   ushort:Union[np.dtype, Callable]=cltypes.ushort
   uint:Union[np.dtype, Callable]=cltypes.uint
   ulong:Union[np.dtype, Callable]=cltypes.ulong
   half:Union[np.dtype, Callable]=cltypes.half
   float:Union[np.dtype, Callable]=cltypes.float
   double:Union[np.dtype, Callable]=cltypes.double

@dataclass(frozen=True)
class _ClTypes(ClTypesScalar, ClTypesVector):
   pass



from pyopencl_extension import Types, use_existing_file_for_emulation, Fft, IFft, to_device


import time
from pytest import mark, fixture
import numpy as np


def get_in_data_cplx(M, N, dtype=Types.cdouble):
    np.random.seed(0)
    in_data_np = np.random.random((M, N,)).astype(dtype)
    in_data_np.imag = np.random.random((M, N,)).astype(in_data_np.imag.dtype)
    return in_data_np


def get_in_data_cplx_radix(radix=2, exponent=10):
    np.random.seed(0)
    return get_in_data_cplx(50, radix ** exponent)


@fixture(params=[
    # np.random.random((2 ** 14,)).astype(Types.cdouble),
    get_in_data_cplx(1, 512, dtype=Types.cdouble),
    get_in_data_cplx(20, 10000, dtype=Types.cdouble),
    # np.random.random((1,2 ** 9,)).astype(Types.cdouble),
    get_in_data_cplx_radix(radix=16, exponent=3),
    get_in_data_cplx_radix(radix=2, exponent=10),
    get_in_data_cplx_radix(radix=4, exponent=5),
    get_in_data_cplx_radix(radix=8, exponent=4)])
def in_data_np(request):
    return request.param


def test_fft(in_data_np):
    atol = 1e-4 if in_data_np.dtype == Types.cfloat else 1e-8
    import numpy as np

    in_data_cl = to_device(in_data_np)

    fft_cl = Fft(in_data_cl, emulate=False)

    # zero padding data for numpy
    axis = 1
    N = in_data_np.shape[axis]
    if not np.log2(N).is_integer():  # if not power of 2, pad accordingly
        N = 2 ** int(np.log2(N) + 1)
    in_data_np_power_of_two = np.zeros((in_data_np.shape[0], N), in_data_np.dtype)
    in_data_np_power_of_two[:, :in_data_np.shape[axis]] = in_data_np

    def measure(call):
        attempts = 3
        ts = []
        for i in range(attempts):
            t1 = time.time()
            call()
            t2 = time.time()
            ts.append(t2 - t1)
        return min(ts)

    # import pyfftw
    # t_fftw = measure(lambda: pyfftw.interfaces.numpy_fft.fft(in_data_np_power_of_two, axis=-1))
    t_np = measure(lambda: np.fft.fft(in_data_np_power_of_two, axis=-1))
    fft_in_data_np = np.fft.fft(in_data_np_power_of_two, axis=-1)

    def fft_call():
        fft_in_data_cl = fft_cl()
        fft_in_data_cl.queue.finish()

    t_cl = measure(fft_call)
    fft_in_data_cl = fft_cl()

    if in_data_np.size < 1024:
        # Test against emulation (commented since it is slower)
        use_existing_file_for_emulation(False)
        fft_cl_py = Fft(in_data_cl, emulate=True)

        fft_in_data_cl_py = fft_cl_py()
        a = fft_in_data_cl_py.get().view(Types.cdouble)
        b = fft_in_data_cl.get().view(Types.cdouble)
        c = fft_in_data_np.view(Types.cdouble)
        assert np.allclose(a, b)
        assert np.allclose(c, b)
        assert np.allclose(c, a)

    # import matplotlib.pyplot as plt
    # plt.plot(fft_in_data_np.flatten())
    # plt.plot(fft_in_data_cl_emulation.get().flatten())
    # plt.show()
    assert np.allclose(fft_in_data_np, fft_in_data_cl.get(), atol=atol)
    # benchmark using reikna
    if False:  # change to true to run against reikna's fft. Note: Reikna takes quite some optimization time before run
        from reikna.cluda import any_api
        from reikna.fft import FFT
        import numpy
        api = any_api()
        thr = api.Thread.create()
        data = in_data_np
        dtype = data.dtype
        axes = (1,)
        fft = FFT(data, axes=axes)
        fftc = fft.compile(thr)
        data_dev = thr.to_device(data)
        res_dev = thr.empty_like(data_dev)
        ts = []
        for i in range(attempts):
            t1 = time.time()
            fftc(res_dev, data_dev)
            thr.synchronize()
            t2 = time.time()
            ts.append(t2 - t1)
        fwd_ref = numpy.fft.fftn(data, axes=axes).astype(dtype)
        tnp = time.time()
        fwd_ref = numpy.fft.fftn(data, axes=axes).astype(dtype)
        tnp = time.time() - tnp
        # numpy.fft.fftn(data[:, :, 0], axes=(1,))
        treikna_min = min(ts)
        assert np.allclose(fft_in_data_np, res_dev.get())


def test_ifft(in_data_np):
    in_ = to_device(in_data_np)
    fft_cl = Fft(in_buffer=in_)
    fft_in_ = fft_cl()

    ref_ifft_fft_in_ = np.fft.ifft(fft_in_.get(), axis=-1)
    ifft_cl = IFft(fft_in_)
    ifft_fft_in_ = ifft_cl().get()
    if in_data_np.size < 1024:
        ifft_cl_py = IFft(fft_in_, emulate=True)
        py_np_ifft_fft_in_ = ifft_cl_py().get()
        a = py_np_ifft_fft_in_
        b = ifft_fft_in_
        assert np.allclose(a, b)
    assert np.allclose(ifft_fft_in_, ref_ifft_fft_in_)



from .modifications_pyopencl.context import Context, get_devices
from .modifications_pyopencl.command_queue import Profiling, CommandQueue, create_queue, get_device, \
    set_current_queue, get_current_queue, set_default_device, activate_profiling, evaluate_profiling
from .modifications_pyopencl.array import zeros, zeros_like, empty, Array, empty_like, to_device
from .emulation import use_existing_file_for_emulation
from .framework import (Private, Scalar, Local, Global, Constant, Function,
                        Kernel,
                        Program, get_vec_size, LocalArray, HashArray, Helpers, Types,
                        scalar_type_from_vec_type, KernelGridType, int_safe, create_cl_files)
from .helpers.general import (typed_partial)
from .types.utilities_np_cl import (match_vec_size, c_name_from_dtype, is_complex_type, is_vector_type, VEC_INDICES,
                                    match_integer_type_for_select)
from .types.auto_gen.cl_types_import import *
from .components.copy_array_region import (CopyArrayRegion, Slice, cl_set)
from .components.fft import (Fft, IFft, FftStageBuilder)
from .components.sumalongaxis import SumAlongAxis
from .components.transpose import Transpose
from .modifications_pyopencl.command_queue import QueueProperties


from typing import Tuple

from pyopencl_extension import Kernel, Global, Array, empty

__author__ = "P.Mohr"
__copyright__ = "26.03.2020, P.Mohr"
__version__ = "1.0"
__email__ = "philipp.mohr@tuhh.de"
__doc__ = """This module contains the Transpose class which transposes an array."""

"""
/*
in_buffer.shape =(4,8,10)
axis_order = (1,0,2)
out_buffer.shape = (8,4,10)

s0 = get_global_size(1) = 4
s1 = get_global_size(1) = 8
s2 = get_global_size(2) = 10

Computation grid is spanned by in_buffer.shape:
g0,g1,g2

According to axis order, elements of in_buffer are copied to different positions in memory of out_buffer:
for axis_order = (1,0,2):
out_buffer[g1*s0*s2+g0*s2+g2]=in_buffer[g0*s1*s2+g1*s2+g2]
i_in = g0*s1*s2+g1*s2+g2
i_out = g1*s0*s2+g0*s2+g2

1 D scenario for axis_order = (1,0):
s0 = get_global_size(1) = 4
s1 = get_global_size(1) = 1

i_in = g0*s1+g1
i_out = g1*s0+g0

*/
"""


class Transpose:
    def _command_for_input_address_computation(self):
        """
        //specific implementation axes_order=(1,0)
        int i_in = g0*get_global_size(1)+g1;
        int i_out = g1*get_global_size(0)+g0;

        //specific implementation axes_order=(1,2,0)
        int i_in = g0*get_global_size(1)*get_global_size(2)+g1*get_global_size(2)+g0;
        int i_out = g1*get_global_size(2)*get_global_size(0)+g2*get_global_size(0)+g0;
        """
        n_dim = self.in_buffer.shape.__len__()
        command = '0'
        for i in range(n_dim):
            offset = '1'
            for j in range(i + 1, n_dim):
                offset += '*get_global_size({})'.format(j)
            command += '+get_global_id({})*{}'.format(i, offset)
        return command  # 'get_global_id(0)*get_global_size(1)+get_global_id(1)'

    def _command_for_output_address_computation(self):
        """
        //specific implementation axes_order=(1,0)
        int i_in = g0*get_global_size(1)+g1;
        int i_out = g1*get_global_size(0)+g0;

        //specific implementation axes_order=(1,2,0)
        int i_in = g0*get_global_size(1)*get_global_size(2)+g1*get_global_size(2)+g0;
        int i_out = g1*get_global_size(2)*get_global_size(0)+g2*get_global_size(0)+g0;
        """
        n_dim = self.in_buffer.shape.__len__()
        command = '0'
        for i in range(n_dim):
            offset = '1'
            # axes_order[j]: What has been the axis of input buffer related to axis j of output buffer?
            for j in range(i + 1, n_dim):
                offset += '*get_global_size({})'.format(self.axes_order[j])
            command += '+get_global_id({})*{}'.format(self.axes_order[i], offset)
        return command  # 'get_global_id(1)*get_global_size(0)+get_global_id(0)'

    def __init__(self, in_buffer: Array, axes_order: Tuple[int, ...]):
        self.axes_order = axes_order
        shape_out = tuple([list(in_buffer.shape)[i] for i in axes_order])
        self.out_buffer = empty(shape_out, in_buffer.dtype)
        self.in_buffer = in_buffer
        self.knl = Kernel(name='transpose',
                          args={'in_buffer': Global(self.in_buffer, read_only=True),
                                'out_buffer': Global(self.out_buffer)},
                          body=["""
                                int i_in = ${i_in};
                                int i_out = ${i_out};
                                out_buffer[i_out] = in_buffer[i_in];                       
                               """],
                          replacements={'i_in': self._command_for_input_address_computation(),
                                        'i_out': self._command_for_output_address_computation()},
                          global_size=self.in_buffer.shape).compile()

    def __call__(self, *args, **kwargs):
        self.knl()
        return self.out_buffer



from typing import Tuple, Any

import numpy as np

from pyopencl_extension.types.auto_gen.cl_types import ClTypesScalar
from pyopencl_extension.types.utilities_np_cl import c_name_from_dtype, dtype_from_c_name, VEC_INDICES, \
    VEC_INDICES_XYZW, Types, VEC_ADDRESSING, scalar_type_from_vec_type
import pyopencl_extension.modifications_pyopencl.cltypes as tp


class VecVal:
    """
    This class is a container for simulating the behavior of OpenCl vector types when using Python and Numpy.

    E.g. when adding two number a and b of type float2, the addition operation is not implemented by default.
    THe implementation of addition operation is included in this class: Each element of the all vector elements of
    a and b are added separately to a new number which is then return as another instance of VecVal.

    Likewise other operation are implemented in this class.
    """

    def __init__(self, val):
        self.val = val
        self.type_handler = TypeHandlerVec(c_name_from_dtype(val.dtype))
        self.vec_size = len(val.dtype.descr)

    def _perform_on_each_element(self, other, operation):
        # e.g. 1) vec[0] *= (a, b); instead of 2) vec[0] *= (vectype)(a, b);
        # case 1) leads to erroneous behavior in opencl, therefore therefore error
        if not isinstance(other, VecVal):
            # other = self.type_handler(*other)
            raise ValueError('Input type is invalid')

        res = tuple([getattr(self.val[f's{i}'], operation)(other.val[f's{i}'])
                     for i in range(self.vec_size)])
        return self.type_handler(*res)

    vec_indices = [f's{idx}' for idx in VEC_INDICES] + VEC_INDICES_XYZW
    vec_special = VEC_ADDRESSING
    scalar_types_builtins = [int, float]
    scalar_types_cl = list(ClTypesScalar().__dict__.values())
    scalar_types = scalar_types_builtins + scalar_types_cl

    def __setattr__(self, name: str, value: Any) -> None:
        if name in self.vec_indices:
            self.val[name] = value
        elif name in self.vec_special and type(value) in self.scalar_types:
            for i in self.vec_special[name][str(len(self.val.dtype.names))]:
                self.val[f's{i}'] = value
        elif name in self.vec_special and type(value) == VecVal:
            raise NotImplementedError()
        else:
            super().__setattr__(name, value)

    def __getattr__(self, item):
        # https://stackoverflow.com/questions/2405590/how-do-i-override-getattr-in-python-without-breaking-the-default-behavior
        if item in self.vec_indices:
            return self.val[f'{item}']
        elif item in VEC_ADDRESSING.keys():
            # get type
            items = [self.val[f's{i}'] for i in VEC_ADDRESSING[item][f'{self.vec_size}']]
            typename = c_name_from_dtype(scalar_type_from_vec_type(self.val.dtype))
            if len(items) > 1:
                typename += str(len(items))
                return TypeHandlerVec(typename)(*items)
            else:
                return TypeHandlerScalar(typename)(*items)
        else:
            raise AttributeError

    # Alternative to intercept all magic methods:
    # https://stackoverflow.com/questions/9057669/how-can-i-intercept-calls-to-pythons-magic-methods-in-new-style-classes
    def __add__(self, other):
        return self._perform_on_each_element(other, '__add__')

    def __mul__(self, other):
        return self._perform_on_each_element(other, '__mul__')

    def __sub__(self, other):
        return self._perform_on_each_element(other, '__sub__')

    def __truediv__(self, other):
        return self._perform_on_each_element(other, '__truediv__')

    # todo: implement other build in when required
    def __abs__(self, *args, **kwargs):  # real signature unknown
        """ abs(self) """
        raise NotImplementedError()

    def __and__(self, *args, **kwargs):  # real signature unknown
        """ Return self&value. """
        raise NotImplementedError()

    def __bool__(self, *args, **kwargs):  # real signature unknown
        """ self != 0 """
        raise NotImplementedError()

    def __ceil__(self, *args, **kwargs):  # real signature unknown
        """ Ceiling of an Integral returns itself. """
        raise NotImplementedError()

    def __divmod__(self, *args, **kwargs):  # real signature unknown
        """ Return divmod(self, value). """
        raise NotImplementedError()

    def __eq__(self, *args, **kwargs):  # real signature unknown
        """ Return self==value. """
        raise NotImplementedError()

    def __ge__(self, *args, **kwargs):  # real signature unknown
        """ Return self>=value. """
        raise NotImplementedError()

    def __gt__(self, *args, **kwargs):  # real signature unknown
        """ Return self>value. """
        raise NotImplementedError()

    def __hash__(self, *args, **kwargs):  # real signature unknown
        """ Return hash(self). """
        raise NotImplementedError()

    def __index__(self, *args, **kwargs):  # real signature unknown
        """ Return self converted to an integer, if self is suitable for use as an index into a list. """
        raise NotImplementedError()

    def __int__(self, *args, **kwargs):  # real signature unknown
        """ int(self) """
        raise NotImplementedError()

    def __invert__(self, *args, **kwargs):  # real signature unknown
        """ ~self """
        raise NotImplementedError()

    def __le__(self, *args, **kwargs):  # real signature unknown
        """ Return self<=value. """
        raise NotImplementedError()

    def __lshift__(self, *args, **kwargs):  # real signature unknown
        """ Return self<<value. """
        raise NotImplementedError()

    def __lt__(self, *args, **kwargs):  # real signature unknown
        """ Return self<value. """
        raise NotImplementedError()

    def __mod__(self, *args, **kwargs):  # real signature unknown
        """ Return self%value. """
        raise NotImplementedError()

    def __neg__(self, *args, **kwargs):  # real signature unknown
        """ -self """
        raise NotImplementedError()

    def __ne__(self, *args, **kwargs):  # real signature unknown
        """ Return self!=value. """
        raise NotImplementedError()

    def __or__(self, *args, **kwargs):  # real signature unknown
        """ Return self|value. """
        raise NotImplementedError()

    def __pos__(self, *args, **kwargs):  # real signature unknown
        """ +self """
        raise NotImplementedError()

    def __pow__(self, *args, **kwargs):  # real signature unknown
        """ Return pow(self, value, mod). """
        raise NotImplementedError()

    def __radd__(self, *args, **kwargs):  # real signature unknown
        """ Return value+self. """
        raise NotImplementedError()

    def __rand__(self, *args, **kwargs):  # real signature unknown
        """ Return value&self. """
        raise NotImplementedError()

    def __rdivmod__(self, *args, **kwargs):  # real signature unknown
        """ Return divmod(value, self). """
        raise NotImplementedError()

    def __repr__(self, *args, **kwargs):  # real signature unknown
        """ Return repr(self). """
        raise NotImplementedError()

    def __rfloordiv__(self, *args, **kwargs):  # real signature unknown
        """ Return value//self. """
        raise NotImplementedError()

    def __rlshift__(self, *args, **kwargs):  # real signature unknown
        """ Return value<<self. """
        raise NotImplementedError()

    def __rmod__(self, *args, **kwargs):  # real signature unknown
        """ Return value%self. """
        raise NotImplementedError()

    def __rmul__(self, *args, **kwargs):  # real signature unknown
        """ Return value*self. """
        raise NotImplementedError()

    def __ror__(self, *args, **kwargs):  # real signature unknown
        """ Return value|self. """
        raise NotImplementedError()

    def __round__(self, *args, **kwargs):  # real signature unknown
        """
        Rounding an Integral returns itself.
        Rounding with an ndigits argument also returns an integer.
        """
        raise NotImplementedError()

    def __rpow__(self, *args, **kwargs):  # real signature unknown
        """ Return pow(value, self, mod). """
        raise NotImplementedError()

    def __rrshift__(self, *args, **kwargs):  # real signature unknown
        """ Return value>>self. """
        raise NotImplementedError()

    def __rshift__(self, *args, **kwargs):  # real signature unknown
        """ Return self>>value. """
        raise NotImplementedError()

    def __rsub__(self, *args, **kwargs):  # real signature unknown
        """ Return value-self. """
        raise NotImplementedError()

    def __rtruediv__(self, *args, **kwargs):  # real signature unknown
        """ Return value/self. """
        raise NotImplementedError()

    def __rxor__(self, *args, **kwargs):  # real signature unknown
        """ Return value^self. """
        raise NotImplementedError()

    def __sizeof__(self, *args, **kwargs):  # real signature unknown
        """ Returns size in memory, in bytes. """
        raise NotImplementedError()

    def __trunc__(self, *args, **kwargs):  # real signature unknown
        """ Truncating an Integral returns itself. """
        raise NotImplementedError()

    def __xor__(self, *args, **kwargs):  # real signature unknown
        """ Return self^value. """
        raise NotImplementedError()


class TypeHandlerScalar:
    """
    This class types scalar OpenCl types.
    """

    def __init__(self, dtype):
        self.dtype = dtype

    def __call__(self, arg):
        from pyopencl_extension.types.funcs_for_emulation import CPointer
        if isinstance(arg, CPointer):
            return arg.view(self.dtype)
        else:
            return np.dtype(self.dtype).type(arg)


class TypeHandlerVec:
    """
    This class types vector OpenCl types.
    """

    def __init__(self, name):
        self.dtype = dtype_from_c_name(name)
        self._type = eval(f'tp.make_{name}')

    def __call__(self, *args):
        from pyopencl_extension.types.funcs_for_emulation import CPointer
        from pyopencl_extension.types.funcs_for_emulation import CPointerVec
        if isinstance(args[0], CPointer) or isinstance(args[0], CPointerVec):
            return args[0].view(self.dtype)
        elif isinstance(args[0], VecVal):
            return args[0]
        elif len(args) == 1:
            a = self._type(*(args * len(self.dtype.names)))
            return VecVal(a)
        elif isinstance(args, tuple):
            a = self._type(*args)
            return VecVal(a)
        else:
            raise ValueError(f'Input not supported: {args=}')


def test_vec_val():
    import auto_gen.types_for_emulation as tp
    a = tp.long2(2, 2)
    a.even = 1
    assert a.val == tp.long2(1, 2).val
    a.odd = 3
    assert a.val == tp.long2(1, 3).val
    a.s0 = 3
    assert a.val == tp.long2(3, 3).val

    a = tp.long2(1, 2)
    b = tp.long2(1, 2)
    assert (a + b).val == tp.long2(2, 4).val
    assert (a - b).val == tp.long2(0, 0).val
    assert (a * b).val == tp.long2(1, 4).val
    assert tp.long2(1).val == tp.long2(1, 1).val


def test_vec_val_and_carray():
    from pyopencl_extension.types.funcs_for_emulation import CPointerVec
    from pyopencl_extension import Types
    import pyopencl_extension.types.auto_gen.types_for_emulation as tp_emulation
    a = tp_emulation.long2(1, 2)
    b = tp_emulation.long2(1, 2)

    ary = CPointerVec(10, Types.long2)
    ary[0] = a + b
    assert ary[0].val == tp_emulation.long2(2, 4).val

    ary = CPointerVec(10, Types.long2)
    ary_typed = tp_emulation.long4(ary)
    assert ary_typed.memory.dtype == Types.long4



import subprocess
from pathlib import Path


def install(path_pyopencl):
    path_pip = Path(os.sys.executable).parent.joinpath('pip')
    if path_pip.exists():
        command = '{} install "{}"'.format(str(path_pip), str(path_pyopencl))
    else:
        path_python = Path(os.sys.executable).parent.joinpath('Scripts/pip')
        command = '{} install "{}"'.format(str(path_python), str(path_pyopencl))
    subprocess.call(command)


if __name__ == '__main__':
    import os

    cwd = os.getcwd()

    install(Path(cwd).joinpath('pyopencl-2021.1.1+cl21-cp38-cp38-win_amd64.whl'))



import numpy as np

from pyopencl_extension import Transpose, to_device, Types


def test_transpose_single_dimension():
    ary_np = np.array([[1, 2, 3]], dtype=Types.int).T
    in_cl = to_device(ary_np)
    transpose_on_input_buffer = Transpose(in_cl, (1, 0))
    out_cl = transpose_on_input_buffer()
    assert np.all(out_cl.get() == ary_np.T)


def test_transpose_int():
    ary_np = np.array([[1, 2, 3], [4, 5, 6]], dtype=Types.int)
    ary_np = np.hstack([ary_np, ary_np])  # array([[1, 2, 3], [4, 5, 6]])
    in_cl = to_device(ary_np)
    transpose_in_buffer = Transpose(in_cl, (1, 0))
    out_cl = transpose_in_buffer()
    assert np.all(out_cl.get() == ary_np.T)


def test_transpose_float():
    ary_np = np.array([[1, 2, 3], [4, 5, 6]], dtype=Types.float)
    ary_np = np.hstack([ary_np, ary_np])  # array([[1, 2, 3], [4, 5, 6]])
    in_cl = to_device(ary_np)
    transpose_in_buffer = Transpose(in_cl, (1, 0))
    out_cl = transpose_in_buffer()
    assert np.all(out_cl.get() == ary_np.T)


def test_transpose_complex():
    ary_np = np.array([[1, 2, 3], [4, 5, 6]], dtype=Types.cfloat)
    ary_np = np.hstack([ary_np, ary_np])  # array([[1, 2, 3], [4, 5, 6]])
    in_cl = to_device(ary_np)
    transpose_in_buffer = Transpose(in_cl, (1, 0))
    out_cl = transpose_in_buffer()
    assert np.all(out_cl.get() == ary_np.T)


import logging
import time

import numpy as np

from pyopencl_extension import Helpers, Program, Kernel, Function, Scalar, Global, HashArray, Array, zeros, \
    to_device
from pyopencl_extension import get_current_queue, set_current_queue, create_queue
from pyopencl_extension.modifications_pyopencl.command_queue import QueueProperties
from pyopencl_extension.types.utilities_np_cl import c_name_from_dtype, Types
import pytest


def test_current_queue_feature_1():
    # whenever create_thread() is called this is set as the current thread. Be careful, that Thread() leads to a new
    # context/queue.
    queue = create_queue()
    assert hash(get_current_queue()) == hash(queue)


def test_current_queue_feature_2():
    set_current_queue(None)
    queue1 = get_current_queue()
    # use get_devices() to get list with available_devices where index corresponds to device id
    set_current_queue(queue1)
    thread1_reused = get_current_queue()
    assert hash(queue1) == hash(thread1_reused)
    set_current_queue(create_queue(device_id=0))
    thread2 = get_current_queue()
    assert hash(queue1) != hash(thread2)
    set_current_queue(queue1)
    thread1_reused = get_current_queue()
    assert hash(queue1) == hash(thread1_reused)


class MyComponentAutomaticArgs:
    def __init__(self):
        self.buff = zeros((10,), Types.short)
        self.knl = Kernel('some_operation',
                          {'buff': Global(self.buff),
                           'number': Scalar(Types.short(3))},
                          ["""
                                buff[get_global_id(0)] = number;
                                """],
                          global_size=self.buff.shape).compile()

    def __call__(self, *args, **kwargs):
        self.knl()


def test_automatic_kernel_arguments():
    component = MyComponentAutomaticArgs()

    component()

    buff_cl = component.buff.get()

    assert np.all(buff_cl == 3.0)


class MyComponentManualArgs:
    def __init__(self, b_create_kernel_file: bool = True):
        self.buff = zeros((10,), Types.short)
        self.knl = Kernel('some_operation',
                          {'buff': Global(self.buff),
                           'number': Scalar(Types.short(1))},
                          ["""
                                buff[get_global_id(0)] = number;
                                """],
                          global_size=self.buff.shape).compile()

    def __call__(self, number: float = 0, **kwargs) -> Array:
        self.knl(number=number, **kwargs)
        return self.buff


def test_manual_kernel_arguments():
    component = MyComponentManualArgs()

    component(number=3)

    buff_cl = component.buff.get()

    assert np.all(buff_cl == 3)


class MyComponentManualNoSuperCall(MyComponentManualArgs):
    def __call__(self, number: float = 0, **kwargs) -> Array:
        self.knl(global_size=self.buff.shape,
                 buff=self.buff,
                 number=number)
        return self.buff


def test_manual_kernel_arguments_no_super_call():
    component = MyComponentManualNoSuperCall()

    component(number=3)

    buff_cl = component.buff.get()

    assert np.all(buff_cl == 3.0)


class MyComponentComplexExample:
    def __call__(self, *args, **kwargs):
        if self.mode == 'a':
            self.program.some_operation(number=1)
        elif self.mode == 'b':
            self.knl(number=1)
        return self.buff

    def __init__(self, mode='a', b_create_kernel_file: bool = True):
        self.mode = mode
        self.buff = zeros((10,), Types.short)
        self.data_t = self.buff.dtype
        func = Function('plus_one',
                        {'buffer': Global(self.buff.dtype),
                         'idx': Scalar(Types.int)},
                        ["""
                                  return buffer[idx]+${some_integer};
                                  """],
                        {'some_integer': 5},
                        returns=Types.float)
        knl = Kernel('some_operation',
                     {'buff': Global(self.buff),
                      'number': Scalar(Types.short(3.0))},
                     ["""
                                data_t factor = convert_${data_t}(1.3);
                                buff[get_global_id(0)] = plus_one(buff, get_global_id(0)) + SOME_CONSTANT*factor;
                                """],
                     replacements={'data_t': c_name_from_dtype(self.data_t)},
                     global_size=self.buff.shape)
        defines = {'SOME_CONSTANT': 6}
        type_defs = {'data_t': self.data_t}
        self.program = Program(defines=defines, type_defs=type_defs, functions=[func],
                               kernels=[knl]).compile()
        self.knl = knl


def test_complex_example():
    component = MyComponentComplexExample()
    component_b = MyComponentComplexExample(mode='b')
    res = component().get()
    res_b = component_b().get()

    assert np.all(res == 11)


def test_complex_example_conversion_python():
    component = MyComponentComplexExample()
    res_cl = component().get()
    res_py = component(emulate=True).get()

    assert np.all(res_cl == res_py - res_cl)


def test_non_existing_argument_raises_warning():
    component = MyComponentComplexExample()
    try:
        res_cl = component(buffer2='something').get()
    except ValueError as err:
        assert str(
            err) == 'keyword argument [\'buffer2\'] does not exist in kernel argument list [\'buff\', \'number\']'


def test_memoize_kernel():
    # thread = Thread(profile=True)
    ary_a = np.ones(int(1e3))
    ary_b = np.zeros(ary_a.shape)
    ary_a_buffer = to_device(ary_a)
    ary_b_buffer = to_device(ary_b)
    n_recompilations = 100
    for i in range(n_recompilations + 1):
        kernels = []
        for j in range(10):
            some_knl = Kernel(f'some_knl_{j}',
                              {'ary_a': Global(ary_a_buffer),
                               'ary_b': Global(ary_b_buffer)},
                              """
                     ary_b[get_global_id(0)] = ary_a[get_global_id(0)];
                     """)
            kernels.append(some_knl)
        Program(kernels=kernels).compile()
        some_knl(global_size=ary_a.shape)
        if i == 1:
            t = time.time()
    time_per_recompile = (time.time() - t) / n_recompilations
    # thread.queue.get_profiler().show_histogram_cumulative_kernel_times()
    print(time_per_recompile)
    assert time_per_recompile < 0.001  # less than 1 ms overhead per recompilation achieved through caching


def test_get_refreshed_argument_of_memoized_kernel():
    for i in range(10):
        ary_a = np.ones(100) + i
        ary_b = np.zeros(100)
        some_knl = Kernel('some_knl',
                          {'ary_a': Global(to_device(ary_a)),
                           'ary_b': Global(to_device(ary_b))},
                          """
                 ary_b[get_global_id(0)] = ary_a[get_global_id(0)];
                 """).compile()
        some_knl(global_size=ary_a.shape)
    assert np.all(some_knl.ary_b.get() == ary_a)


def test_kernel_arg_type_conversion():
    mem = {'ary_b': zeros(shape=(100,), dtype=Types.int)}
    for i in range(5):
        ary_a = np.ones(100, Types.int)
        some_knl = Kernel('some_knl',
                          mem | {'ary_a': ary_a,
                                 'offset': float(i),  # checks if float is accepted
                                 'val': Types.ushort(5.0)  # just a dummy value to test if ushort is accepted
                                 },
                          'ary_a[get_global_id(0)] = ary_a[get_global_id(0)] + offset;' + \
                          'ary_b[get_global_id(0)] = ary_b[get_global_id(0)] + offset;',
                          global_size=ary_a.shape).compile()
        some_knl()
    assert np.all(some_knl.ary_a.get() == ary_a + 4)  # every kernel call the numpy array is send to device
    assert np.all(10 == mem['ary_b'].get())


logging.basicConfig(level=logging.INFO)


def test_local_from_global_dimenstions():
    local_size = Helpers._get_local_size_coalesced_last_dim(global_size=(10000, 128), desired_wg_size=64)
    assert local_size == (1, 64)
    local_size = Helpers._get_local_size_coalesced_last_dim(global_size=(13, 13), desired_wg_size=64)
    assert local_size == (1, 13)
    local_size = Helpers._get_local_size_coalesced_last_dim(global_size=(10000,), desired_wg_size=64)
    assert local_size == (50,)


def test_multiple_command_queues():
    queue1 = create_queue(device_id=0)
    queue2 = create_queue(context=queue1.context)
    ary_a = to_device(np.ones(100000) + 1, queue1)
    ary_b = to_device(np.zeros(100000), queue1)
    some_knl = Kernel('some_knl',
                      {'ary_a': Global(ary_a),
                       'ary_b': Global(ary_b)},
                      """
             ary_b[get_global_id(0)] += ary_a[get_global_id(0)];
             """, global_size=ary_a.shape).compile(queue2.context)
    some_knl(queue=queue2)
    # thread2.queue.finish()
    some_knl(queue=queue1)
    test = 0


def test_hash_array():
    ary = zeros(shape=(100,), dtype=Types.float)
    hash_ary = HashArray(ary)
    a_hash = hash_ary.hash.copy()
    hash_ary.set(np.ones(hash_ary.shape).astype(hash_ary.dtype))
    b_hash = hash_ary.hash.copy()
    assert a_hash != b_hash
    hash_ary[0] = 5
    c_hash = hash_ary.hash
    assert c_hash != b_hash


# to much work to make overloading functionality working from outside of C compiler, because that requires tracking
# types of variables. If complex support is required just make second implementation.
# def test_real_complex_support():
#     thread = ClInit()
#     ary_a = to_device(thread.queue, np.ones(100))
#     ary_b = to_device(thread.queue, np.ones(100))
#     knl = ClKernel('sum_and_multiply',
#                    {'a': KnlArgBuffer(),
#                     'b': KnlArgBuffer()},
#                    """
#              int i = get_global_id(0);
#              a[i] = (a[i]+3) *b[i];
#              """).compile()

@pytest.mark.skip()
def test_profiling():
    # todo: python (net) time seems not correct
    # - How to deal with multiple events created in e.g. zeros()
    queue = create_queue(queue_properties=QueueProperties.PROFILING_ENABLE)
    size = int(1e8)
    for i in range(10):
        _ = zeros((size,), dtype=Types.int)
    get_current_queue().get_profiler().show_histogram_cumulative_kernel_times()
    queue.finish()


# def test_kernel_timeit():
#     # todo: python (net) time seems not correct
#     # - How to deal with multiple events created in e.g. zeros()
#     thread = Thread(profile=True)
#     ary_a = np.ones(100) + 5
#     ary_b = np.zeros(100)
#     some_knl = Kernel('some_knl',
#                       {'ary_a': Global(to_device(thread.queue, ary_a)),
#                        'ary_b': Global(to_device(thread.queue, ary_b))},
#                       """
#              ary_b[get_global_id(0)] = ary_a[get_global_id(0)];
#              """, global_size=ary_a.shape).compile(thread)
#     some_knl.timeit(reps=1, number=10)


def test_add_functions_inside_function_or_kernel_definition():
    ary_a = to_device(np.ones(100))
    fnc_add3 = Function('add_three',
                        {'a': Scalar(Types.int)},
                        'return a + 3;', returns=Types.int)
    fnc_add5 = Function('add_five',
                        {'a': Scalar(Types.int)},
                        """
             return add_three(a)+2;
             """,
                        functions=[fnc_add3], returns=Types.int)
    some_knl = Kernel('some_knl',
                      {'ary_a': Global(ary_a)},
                      """
             ary_a[get_global_id(0)] = add_five(ary_a[get_global_id(0)]);
             """, global_size=ary_a.shape,
                      functions=[fnc_add5])
    functions = []  # funcitons defined here have higher proiority in case of name conflicts
    Program(functions=functions, kernels=[some_knl]).compile()
    some_knl()
    assert ary_a.get()[0] == 6


def test_conversion_knl_fnc_args_with_no_pointer_format():
    # import pyopencl_extension as cl
    # cl.activate_profiling()
    a_np = np.array([0.1, 0.2], dtype=Types.float)
    b_cl = zeros(shape=(2,), dtype=Types.float)
    fnc = Function('copy_fnc',
                   {'a': a_np, 'b': b_cl, 'idx': Scalar(Types.int)},
                   """
                   b[idx] = a[idx];
                   """)
    knl = Kernel('some_knl',
                 {'a': a_np, 'b': b_cl},
                 """
                 copy_fnc(a, b, get_global_id(0));
                 """, functions=[fnc], global_size=b_cl.shape)
    knl.compile()
    knl()
    # cl.evaluate_profiling()
    assert np.all(a_np == b_cl.get())



import inspect
import logging
import os
import re
from collections import namedtuple
from dataclasses import dataclass, field
from pathlib import Path
from textwrap import indent
from typing import Tuple, List

import numpy as np
from pycparser.c_ast import BinaryOp, ID, IdentifierType, Return, Constant, Assignment, Decl, FuncDef, Node, TypeDecl, \
    FuncCall, ExprList, Compound, ArrayRef, For, If, Cast, ArrayDecl, UnaryOp, While, Break, InitList, Typedef, \
    StructRef, PtrDecl, TernaryOp, EmptyStatement
from pycparserext.ext_c_parser import FuncDeclExt, PreprocessorLine, OpenCLCParser

__author__ = "P.Mohr"
__copyright__ = "26.03.2020, P.Mohr"
__version__ = "1.0"
__email__ = "philipp.mohr@tuhh.de"
__doc__ = """This script includes helpful functions to extended PyOpenCl functionality."""

# make sure emulation does not import anything from framework, to avoid circular imports
# framework does depends on emulation and not the other way around.
from pyopencl._cl import LocalMemory

from pyopencl_extension.types.funcs_for_emulation import init_array, CPointer, CPointerVec
from pyopencl_extension.helpers.general import write_string_to_file
from pyopencl_extension import Array
from pyopencl.array import Array as ClArray

arrays_cls = (Array, ClArray)

# following lines are used to support functions from <pyopencl-complex.h>
from pyopencl_extension.types.utilities_np_cl import is_vector_type

preamble_buff_t_complex_np = lambda cplx_t: """
{cplx_t}_mul = lambda x, y: {cplx_t}_t(x * y)
{cplx_t}_add = lambda x, y: {cplx_t}_t(x + y)
{cplx_t}_sub = lambda x, y: {cplx_t}_t(x - y)
{cplx_t}_abs = lambda x: {cplx_t}_t(np.abs(x))
{cplx_t}_rmul = lambda x_real, y_cplx: {cplx_t}_t(x_real * y_cplx)
{cplx_t}_new = lambda real, imag: {cplx_t}_t(complex(real, imag))
{cplx_t}_conj = lambda x: np.conj(x)
""".format(cplx_t=cplx_t)

preamble_buff_t_complex64_np = 'cfloat_t = np.dtype(\'complex64\').type\nPI = 3.14159265359\n{}'.format(
    preamble_buff_t_complex_np('cfloat'))
preamble_buff_t_complex128_np = 'cdouble_t = np.dtype(\'complex128\').type\nPI = 3.14159265358979323846\n{}'.format(
    preamble_buff_t_complex_np('cdouble'))


# unparse to python
def py_indent(text):
    return indent(text, 4 * ' ')


# This section contains function that raise error messages if pattern likely causing an erroneous behaviour is detected
def detect_inconsistent_for_loop_header_variable(node: For):
    # case: for(int i=0; t<10;t++){}, raises error if variable in decl, cond and incr do not match
    init_var = node.init.decls[0].name
    if isinstance(node.cond.left, ID):
        cond_var = node.cond.left.name
    else:
        cond_var = node.cond.right.name
    incr_var = node.next.expr.name
    if not init_var == cond_var == incr_var:
        raise ValueError(
            f'Line {node.coord.line}: For loop header contains different variables {init_var=} {cond_var=} {incr_var=}')


# @dataclass
# class Container:
#     unparsed_code: str
#     # https://overiq.com/c-programming-101/increment-and-decrement-operators-in-c/
#     code_unary_operator_prefix: List[str] = None  # e.g. array[++x], y= ++x;  array[--x]; y = --x;
#     code_unary_operator_postfix: List[str] = None  # e.g. array[x++], y= x++;  array[x--]; y = x--;

# https://stackoverflow.com/questions/2673651/inheritance-from-str-or-int
class Container(str):
    def __new__(cls, value, code_unary_operator_prefix: List[str] = None,
                code_unary_operator_postfix: List[str] = None):
        obj = str.__new__(cls, value)
        if code_unary_operator_prefix is None:
            code_unary_operator_prefix = []
        obj.code_unary_operator_prefix = code_unary_operator_prefix
        if code_unary_operator_postfix is None:
            code_unary_operator_postfix = []
        obj.code_unary_operator_postfix = code_unary_operator_postfix
        return obj

    def wrap_code_with_prefix_postfix(self) -> str:
        str_ = self
        if not len(self.code_unary_operator_prefix) == 0:
            str_ = '{}\n{}'.format('\n'.join(self.code_unary_operator_prefix), str_)
        if not len(self.code_unary_operator_postfix) == 0:
            str_ = '{}\n{}'.format(str_, '\n'.join(self.code_unary_operator_postfix))
        return str(str_)

    def only_use_prefix_postfix(self):
        """
        In case of a nested unary operation like array[number++]=something; the result in Python should be
        array[number] = something
        number += 1.

        In case of a separate unary operation like number++; the result in Python should be
        number+=1 but not numbernumber+=1 as it would be we handled like in the nested case.

        Therefore, this function allows to extract only the unary strings.
        :return:
        """
        str_ = ''
        if not len(self.code_unary_operator_prefix) == 0:
            str_ = '{}'.format('\n'.join(self.code_unary_operator_prefix))
        if self.code_unary_operator_postfix is not None:
            str_ = '{}'.format('\n'.join(self.code_unary_operator_postfix))
        return str(str_)

    def add_container_pre_post_fix(self, container: 'Container'):
        self.code_unary_operator_postfix.extend(container.code_unary_operator_postfix)
        self.code_unary_operator_prefix.extend(container.code_unary_operator_prefix)


regex_real_imag_assignment = re.compile(r'([^\[]+)(\[)(.+)(\])(\.)(real|imag)')  # #define PI 3.14159265358979323846\n

names_func_require_work_item = []

names_func_has_barrier = []


class MacroWithArguments:
    """ Explanation:
    Consider following macro with arguments
    #define DFT2_TWIDDLE(a,b,t) { data_t tmp = t(a-b); a += b; b = tmp; }
    which is used e.g. as:
    DFT2_TWIDDLE(u[0],u[8],mul_p0q8);

    1. replace_macro_having_argument_with_function(code_c)
    # replace defines with functions
    # e.g. '# define DFT2_TWIDDLE(a,b,t) { data_t tmp = t(a-b); a += b; b = tmp; }'
    # becomes: 'void DFT2_TWIDDLE(a,b,t){ data_t tmp = t(a-b); a += b; b = tmp; }

    2. c-parser converts function to function to abstract syntax tree model

    3. unparse_macro_node_and_convert_to_string:
        In a first step this macro is converted to python code which yields:
        def DFT2_TWIDDLE(a, b, t, wi: WorkItem = None):
            tmp = data_t(t(a - b))
            a += b
            b = tmp

        To achieve the same behavior as the C macro, we have to execute the python function code in the
        scope where the function DFT2_TWIDDLE is called. The variables in the arguments have to be
        substituted accordingly.
        We get:
        DFT2_TWIDDLE = 'tmp = data_t({t}({a} - {b}))\n{a} += {b}\n{b} = tmp'
        which is used in the emulator like in the following:
        exec(DFT2_TWIDDLE.format(a='u[0]', b='u[8]', t='mul_p0q8'))
    """
    names_py_macro = {}  # dictionary with executable string

    @classmethod
    def replace_with_function(cls, code_c):
        cls.names_py_macro.clear()  # make sure that dict is empty
        if search(regex_func_def_multi_line, code_c):
            for _ in re.findall(regex_func_def_multi_line, code_c):
                def_part_code_c = ''.join(_)
                res = search(regex_func_def_multi_line, def_part_code_c)
                func = f'void {res.group(3)}({res.group(5)}){{{res.group(9)}}}'
                code_c = code_c.replace(def_part_code_c, func)

                cls.names_py_macro[res.group(3)] = 'to be filled with python code'
        return code_c

    @classmethod
    def unparse_macro_node_and_convert_to_string(cls, node):
        macro = _unparse(node.body)
        args = [p.name for p in node.decl.type.args.params]
        for arg in args:
            macro = re.sub(r'(?<![\w])' + arg + r'(?![\w])', f'{{{arg}}}', macro)
        cls.names_py_macro[node.decl.name] = args
        return f'{node.decl.name} = """{macro}"""'


def _unparse_header(args, node):
    if isinstance(node.type, TypeDecl):
        name = _unparse(node.type)
    else:
        raise ValueError('Not implemented')
    return 'def {}({}, wi: WorkItem = None):'.format(name, args)


def _unparse_func_header(node):
    args = ', '.join(['{}'.format(p.name) for p in node.args.params])
    return _unparse_header(args, node)


def _unparse_knl_header(node):
    """
    If kernel header includes local memory argument, the underlying datatype is extracted.
    e.g. __kernel some_knl(__local short *local_mem)
    is converted in emulation .py file as:
    @cl_kernel
    def some_knl(local_mem: short):
    Later, the datatype "short" is extracted to type the array local_mem in emulation mode.
    """
    args = ', '.join(['{}'.format(p.name) if not any(q in ['local', '__local'] for q in p.quals)
                      else f'{p.name}: {_unparse(p.type.type.type.names[0])}'
                      for p in node.args.params])
    return _unparse_header(args, node)


def _deal_with_func_def_node(node):
    # check if function is kernel

    if len(node.decl.funcspec) > 0:
        if node.decl.funcspec[0] == '__kernel':
            header = _unparse_knl_header(node.decl.type)
            header = f'@cl_kernel\n{header}'
        else:
            raise ValueError('Func spec not supported')
    else:
        header = _unparse(node.decl.type)

    body = _unparse(node.body)
    final_yield = ''
    if len(node.decl.funcspec) > 0:
        if node.decl.funcspec[0] == '__kernel':
            final_yield = py_indent('yield  # required to model local memory behaviour correctly')
    if final_yield == '':
        function = '{}\n{}'.format(header, py_indent(body))
    else:
        function = '{}\n{}\n{}'.format(header, py_indent(body), final_yield)
    return function


def _deal_with_func_call(node: FuncCall):
    if isinstance(node.name, ID):
        func_name = node.name.name
        # todo: use funcs_for_cl_emulation.py for adding support for all cl functions
        if func_name.startswith('convert_'):
            """
            alternative:
            def convert(value, dtype='int32'):
                return np.dtype(dtype).type(value)
            """
            type_name = re.search(r'(convert_)([\w]+)', node.name.name).group(2)
            res = '{}({})'.format(type_name, _unparse(node.args))
        elif 'cos' == func_name:
            res = 'np.cos({})'.format(_unparse(node.args))
        elif 'sin' == func_name:
            res = 'np.sin({})'.format(_unparse(node.args))
        elif func_name in translation_cl_work_item_functions:
            res = 'wi.{}[{}]'.format(translation_cl_work_item_functions[func_name], _unparse(node.args))
        elif 'barrier' == func_name:
            if isinstance((_ := node.args.exprs[0]), BinaryOp):
                barrier_name = f'{node.args.exprs[0].left.name}|{node.args.exprs[0].right.name}'
            else:
                barrier_name = node.args.exprs[0].name
            res = f'\nwi.scope = locals()  # saves reference to objects in scope for debugging other wi in wg \n' \
                  f'yield  # yield models the behaviour of barrier({barrier_name})\n'
        else:
            if func_name in (_ := MacroWithArguments.names_py_macro):
                args = _unparse(node.args).split(', ')
                format_arg = ', '.join([f'{macro_arg}="{args[i]}"' for i, macro_arg in enumerate(_[func_name])])
                res = f'exec({func_name}.format({format_arg}))'
            elif func_name in names_func_require_work_item:
                res = '{}({}, wi=wi)'.format(_unparse(node.name), _unparse(node.args))
                if func_name in names_func_has_barrier:
                    res = f'(yield from {res})'
            else:
                res = '{}({})'.format(_unparse(node.name), _unparse(node.args))
    else:
        res = '{}({})'.format(_unparse(node.name), _unparse(node.args))
    return res


def _deal_with_compound(node: Compound):
    # compound = [_unparse(block_item) for block_item in node.block_items]
    # loop makes debugging easier. Set breakpoint in loop, then remove BP and step over to get to target line
    compound = []
    for block_item in node.block_items:
        if isinstance(block_item, UnaryOp):  # number++; should translate to number +=1 and not numbernumber+=1
            compound.append(_unparse(block_item).only_use_prefix_postfix())
        else:
            compound.append(_unparse(block_item).wrap_code_with_prefix_postfix())
    res = '\n'.join(compound)
    return res


def _deal_with_arrayref(node: ArrayRef):
    # if isinstance(node.subscript, UnaryOp):  # e.g. array[pos++] = 5; -> pos+=1; \n array[pos] = 5;
    #     unary_op = _unparse(node.subscript)
    #     res = '{}[{}]'.format(_unparse(node.name), node.subscript.expr.name)
    # else:
    array_index = _unparse(node.subscript)
    res = Container('{}[{}]'.format(_unparse(node.name), array_index))
    res.add_container_pre_post_fix(array_index)
    return res


def _deal_with_for(node: For):
    # only for loop considered of style: for(int i=start; i<=stop; i++)
    #                                 or for(int i>=stop; i>=0; i--)
    detect_inconsistent_for_loop_header_variable(node)

    var = node.init.decls[0].name
    right = _unparse(node.cond.right)
    left = _unparse(node.cond.left)
    cond_op = node.cond.op
    if right == var:  # e.g. 5>=i -> i<= 5
        right = left
        left = var
        cond_op = {'!=': '!=',
                   '<=': '>=',
                   '>=': '<=',
                   '>': '<',
                   '<': '>'}[cond_op]
    stop = right
    op = node.next.op
    stop = {'p++': {'!=': stop,
                    '<': stop,
                    '<=': stop + '+1'},
            'p--': {'!=': stop,
                    '>': stop,
                    '>=': stop + '-1'}}[op][cond_op]
    if op == 'p++':
        step = 1
    elif op == 'p--':
        step = -1
    else:
        raise NotImplementedError()
    loop_header = '{var} in range({start}, {stop}, {step})'.format(var=var,
                                                                   start=_unparse(node.init.decls[0].init),
                                                                   stop=stop,
                                                                   step=step)
    loop_body = _unparse(node.stmt)
    res = 'for {}:\n{}'.format(loop_header, py_indent(loop_body))
    return res


def _deal_with_if(node: If):
    if_true_cond = '{}'.format(_unparse(node.cond))
    if_true_body = _unparse(node.iftrue)
    if_true = 'if {}:\n{}'.format(if_true_cond, py_indent(if_true_body))
    if node.iffalse is not None:
        else_body = _unparse(node.iffalse)
        else_ = 'else:\n{}'.format(py_indent(else_body))
        res = '{}\n{}'.format(if_true, else_)
    else:
        res = if_true
    return res


def _deal_with_decl(node: Decl):
    if len(node.quals) > 0 and node.quals[0] == 'private':  # private int indices_states_prior[ALPHABET_SIZE];
        res = _unparse(node.type)
    elif isinstance(node.type, ArrayDecl):
        if node.init is None:  # real_t p_x[2];
            res = _unparse(node.type)
        elif isinstance(node.init, InitList):  # real_t p_x[2]={0.0};
            # todo: if array /var is not inialized assign random values or give warning
            array_decl = _unparse(node.type)
            array_fill = [_unparse(item) for item in node.init.exprs]
            # todo: deal with a[5]={1}->a=[1,0,0,0,0} node.type.dim.value == 1 and len(array_fill) == 1 and int(array_fill) != 0:
            if len(array_fill) == 1 and array_fill[0] in ['0', '0.0']:
                res = '{}\n{}.fill({})'.format(array_decl, node.type.type.declname, array_fill[0])
            else:
                raise ValueError('Currently array initializer with multiple or non zero values not supported')
        else:
            raise ValueError('Not supported case')
    elif node.init is None:
        if isinstance(node.type, PtrDecl):  # e.g. float *x;
            res = f'{node.name} = {node.type.type.type.names[0]}(0)'
        else:  # e.g. float x;
            res = f'{node.name} = {node.type.type.names[0]}(0)'
    else:  # int gid1=get_global_id(0);
        if isinstance(node.type, PtrDecl):
            type_cl = node.type.type.type.names[0]
        else:
            type_cl = node.type.type.names[0]
        res = '{} = {}({})'.format(node.name, type_cl, _unparse(node.init))
    return res


def _deal_with_array_decl(node: ArrayDecl):
    if len(node.type.quals) > 0 and node.type.quals[0] in ['local', '__local']:
        res = "{name} = local_memory(wi, '{name}', lambda: init_array({dim}, {dtype}))".format(
            name=_unparse(node.type),
            dim=_unparse(node.dim),
            dtype=node.type.type.names[0])
    else:
        res = f'{_unparse(node.type)} = init_array({_unparse(node.dim)}, {node.type.type.names[0]})'
    return res


def _deal_with_assignment(node: Assignment):
    left = _unparse(node.lvalue)
    right = _unparse(node.rvalue)
    if search(regex_real_imag_assignment, left):
        groups = search(regex_real_imag_assignment, left).groups()
        res = Container(f'set_{groups[5]}(ary={groups[0]}, idx={groups[2]}, value={right})')
    else:
        res = Container('{} {} {}'.format(left, _unparse(node.op), right))
        res.add_container_pre_post_fix(left)
        res.add_container_pre_post_fix(right)
    return res


def _deal_with_binary_op(node: BinaryOp):
    left = _unparse(node.left)
    right = _unparse(node.right)
    # todo: see test_pointer_arithmetics
    # if isinstance(node.left, ID) and node.op in ['+']:
    #     res = f'{node.left.name}[{node.op}:]'
    # elif isinstance(node.right, ID) and node.op in ['+']:
    #     res = f'{node.right.name}[{node.op}:]'
    if node.op in ['%']:
        res = Container(f'c_modulo({left},{right})')
    else:
        if node.op in ['&&']:  # && not supported in python
            node_op = '&'
        elif node.op in ['||']:
            node_op = 'or'
        else:
            node_op = node.op
        # following lines reduce unnecessary brackets like 1+2 is not translated to (1+2) in Python equivalent
        if isinstance(node.right, BinaryOp):
            right = Container(f'({right})', **right.__dict__)
        if isinstance(node.left, BinaryOp):
            left = Container(f'({left})', **left.__dict__)
        res = Container('{} {} {}'.format(left, node_op, right))
    res.add_container_pre_post_fix(left)
    res.add_container_pre_post_fix(right)
    return res


def _deal_with_unary_op(node: UnaryOp):
    if node.op == 'p++':
        res = Container(node.expr.name, code_unary_operator_postfix=['{} += 1'.format(_unparse(node.expr))])
        # res = '{} += 1'.format(_unparse(node.expr))
    elif node.op == 'p--':
        res = Container(node.expr.name, code_unary_operator_postfix=['{} -= 1'.format(_unparse(node.expr))])
    elif node.op == '++':
        res = Container(node.expr.name, code_unary_operator_prefix=['{} += 1'.format(_unparse(node.expr))])
    elif node.op == '--':
        res = Container(node.expr.name, code_unary_operator_prefix=['{} -= 1'.format(_unparse(node.expr))])
    elif node.op == '!':
        res = f'not {_unparse(node.expr)}'
    else:
        res = '{}{}'.format(node.op, _unparse(node.expr))
    return res


def _unparse(node: Node) -> Container:
    match node:
        case FuncDef():
            res = _deal_with_func_def_node(node)
        case FuncDeclExt():
            res = _unparse_func_header(node)
        case TypeDecl():
            res = node.declname
        case FuncCall():
            res = _deal_with_func_call(node)
        case ExprList():
            res = ', '.join([_unparse(expr) for expr in node.exprs])
        case Compound(): # e.g. body of a function or a for loop
            res = _deal_with_compound(node)
        case ArrayRef():
            res = _deal_with_arrayref(node)
        case For():
            res = _deal_with_for(node)
        case While():
            res = 'while {}:\n{}'.format(_unparse(node.cond), py_indent(_unparse(node.stmt)))
        case Break():
            res = 'break'
        case If():
            res = _deal_with_if(node)
        case Decl():
            res = _deal_with_decl(node)
        case ArrayDecl():
            res = _deal_with_array_decl(node)
        case Assignment():
            res = _deal_with_assignment(node)
        case Constant():
            res = node.value
        case Return():
            res = 'return ' + _unparse(node.expr)
        case BinaryOp():
            res = _deal_with_binary_op(node)
        case ID():
            res = node.name
        case IdentifierType():
            res = ''
        case str():
            res = node
        case Cast():
            type_cl = node.to_type.type.type.names[0]
            res = '{}({})'.format(type_cl, _unparse(node.expr))
        case UnaryOp():
            res = _deal_with_unary_op(node)
        case StructRef():
            res = f'{_unparse(node.name)}.{_unparse(node.field)}'
        case TernaryOp():
            res = f'{_unparse(node.iftrue)} if {_unparse(node.cond)} else {_unparse(node.iffalse)}'
        case EmptyStatement():
            res = 'pass'
        case None:
            res = ''
        case _:
            raise ValueError('Node type not considered')

    if not isinstance(res, Container):
        res = Container(res)
    return res


def unparse_function_node(node: Node) -> str:
    return str(_unparse(node))


# put work item functions in tuples such that they are immediately visible when hovering with cursor in debug mode
translation_cl_work_item_functions = {
    'get_global_id': 'global_id',
    'get_local_id': 'local_id',
    'get_global_size': 'global_size',
    'get_local_size': 'local_size',
    'get_num_groups': 'num_groups',
    'get_group_id': 'group_id',
    'get_work_dim': 'work_dim',
}


@dataclass
class WorkItem:
    global_id: tuple[int]
    global_size: tuple[int]
    local_size: tuple[int]
    local_memory_collection: list[dict]
    _local: dict = field(init=False, default=None)
    _group_id_lin: int = None  #
    _scope: dict = None
    work_items: list['WorkItem'] = None
    local_id: tuple[int] = None
    num_groups: tuple[int] = None
    group_id: tuple[int] = None
    work_dim: int = None

    def __post_init__(self):
        n_dim = self.get_work_dim()
        self.work_dim = n_dim
        self.local_id = tuple([self.get_local_id(i) for i in range(n_dim)])
        self.num_groups = tuple([self.get_num_groups(i) for i in range(n_dim)])
        self.group_id = tuple([self.get_group_id(i) for i in range(n_dim)])
        self._group_id_lin = compute_linear_idx(self.group_id, self.num_groups)

    @property
    def info(self):
        n_dim = len(self.global_size)
        local_id = tuple(self.get_local_id(dim) for dim in range(n_dim))
        return f'global_id={self.global_id} local_id={local_id}'

    def get_vars_in_wg(self, name: str):
        """
        Use case: Consider x is a private scoped array of a work item. Then np.vstack(wi.get_scope_vars_in_wg('x'))
        will build a 2d array where all privates scoped arrays named 'x' of the work items of the work group are
        stacked. This feature facilitates debugging where computations are done from cooperating work items in a work
        group.

        :param name:
        :return:
        """
        return np.vstack([_wi.scope[name] for _wi in self.wi_in_wg])

    @property
    def scope(self):
        # remove circular reference to self
        return {k: v for k, v in self._scope.items() if k != 'wi' or k.startswith('__')}

    @scope.setter
    def scope(self, value):
        self._scope = value

    @property
    def wi_in_wg(self):
        return [wi for wi in self.work_items if wi._group_id_as_tuple == self._group_id_as_tuple]

    @property
    def _group_id_as_tuple(self):
        n_dims = len(self.global_size)
        return tuple(self.get_group_id(dim) for dim in range(n_dims))

    @property
    def local(self) -> dict:
        if self._local is None:
            # Use linear index for work group, to retrieve local memory of particular workgroup.
            # All local memory is modeled by self.local_memory_collection
            self._local = self.local_memory_collection[self._group_id_lin]
        return self._local

    def get_global_id(self, dim):
        return self.global_id[dim]

    def get_global_size(self, dim):
        return self.global_size[dim]

    def get_local_id(self, dim) -> int:
        return self.global_id[dim] % self.local_size[dim]

    def get_local_size(self, dim):
        return self.local_size[dim]

    def get_num_groups(self, dim):
        return int(self.global_size[dim] / self.local_size[dim])

    def get_group_id(self, dim):
        return int(self.global_id[dim] / self.local_size[dim])  # return self.local_size[dim]

    def get_global_offset(self, dim):
        raise NotImplementedError()

    def get_work_dim(self):
        return len(self.global_size)


def local_memory(wi: WorkItem, name: str, lambda_array):
    if name not in wi.local:
        wi.local[name] = lambda_array()
    return wi.local[name]


def cl_kernel(kernel):
    """
    Decorater for kernel function emulated with Python.

    :param kernel:
    :return:
    """

    def wrapper_loop_over_grid(global_size, local_size=None, *args):
        args_python = [arg.get().ravel() if isinstance(arg, arrays_cls) else arg for arg in args]
        num_work_groups = 1
        if local_size is None:
            local_size = global_size
        for i in range(len(global_size)):
            num_work_groups *= int(global_size[i] / local_size[i])
        local_memory_collection = [{} for _ in range(num_work_groups)]

        work_items = [WorkItem(global_id=gid, global_size=global_size, local_size=local_size,
                               local_memory_collection=local_memory_collection)
                      for gid in np.ndindex(global_size)]

        # filter for local memory arguments and initialize local memory for each work group
        _args_python = []
        LocalArg = namedtuple('LocalArg', 'name')
        for position, arg in enumerate(args_python):
            if isinstance(arg, LocalMemory):
                arg_name = (_ := inspect.getfullargspec(kernel)).args[position]
                arg_type = _.annotations[arg_name]
                for loc_mem in local_memory_collection:
                    loc_mem[arg_name] = init_array(size=int(arg.size / np.dtype(arg_type.dtype).itemsize),
                                                   type_c=arg_type)
                _args_python.append(LocalArg(arg_name))
            else:
                _args_python.append(arg)
        args_python = _args_python

        def decide_ary_view(arg):
            if is_vector_type(arg.dtype):
                return CPointerVec.from_np(arg)
            else:
                return CPointer.from_np(arg)

        args_python = [decide_ary_view(arg) if isinstance(arg, np.ndarray) else arg
                       for arg in args_python]
        for wi in work_items:
            wi.work_items = work_items

        # Assign work items to individual work groups. Items of a work group can by synchronized using barriers.
        # However, synchronization between work groups is not possible according to OpenCl standard.
        # https://stackoverflow.com/questions/6890302/barriers-in-opencl
        work_items_per_wg = [[] for _ in range(num_work_groups)]
        for wi in work_items:
            work_items_per_wg[compute_linear_idx(wi.group_id, wi.num_groups)].append(wi)

        def prepare_args(work_item):
            return [work_item.local[arg.name] if isinstance(arg, LocalArg) else arg for arg in args_python]

        for i_wg in range(num_work_groups):
            blocking_kernels = [kernel(*prepare_args(work_item), wi=work_item)
                                for work_item in work_items_per_wg[i_wg]]
            while True:
                try:
                    [next(knl) for knl in blocking_kernels]
                except StopIteration:
                    break
        [arg.set(args_python[idx].memory.reshape(arg.shape)) if isinstance(arg, arrays_cls) else arg
         for idx, arg in enumerate(args)]

    return wrapper_loop_over_grid


# test regex at https://regexr.com/
regex_constant = re.compile(r'#define[ ]+([\w]+)[ ]+([\w\.\-e]+)')  # #define PI 3.14159265358979323846\n
regex_func_def = re.compile(r'#define[ ]+([\w]+)\(([\w, ]+)\)[ ]+([\w\.\(\)*\-+/ ]+)')  # #define MUL(x,y) (x*y)\n

# #define DFT2_TWIDDLE(a,b,t) { data_t tmp = t(a-b); a += b; b = tmp; }
regex_func_def_multi_line = re.compile(r'(#define)([ ]+)([\w]+)(\()([\w, ]+)(\))([ ]+)(\{)(.+)(\})')
regex_include = re.compile(r'#include[ ]+<([\w\-\.]+>)')  # '#include <pyopencl-complex.h>\n
regex_numbers_with_conversion_characater = re.compile(r'([\d.]+)([fd])')  # define PI 3.14159265359f
# define SOME_ARRAY (int[]){0, 1, 2, 3}
regex_array_definition = re.compile(r'#define[ ]+([\w]+)[ ]+\((.+)\[\]\)+\{(.+)\}')

search = lambda regex, text: re.search(regex, text)


def unparse_preprocessor_line(line: PreprocessorLine) -> str:
    """
    contents = '#define PI 3.14159265358979323846\n'
    contents = '#define MUL(x,y) (x*y)\n'
    contents = '#define NUM 1e-6\n'

    :param preprocessor_line:
    :return:
    """
    contents = line.contents

    if res := search(regex_numbers_with_conversion_characater, contents):
        replacement = res.group(1)
        contents = re.sub(regex_numbers_with_conversion_characater, replacement, contents)
    if res := search(regex_constant, contents):
        search_convert_t = re.search(r'(convert_)([\w]+)\((.*)\)', contents)
        if search_convert_t:  # catch case e.g. LLR_MAX convert_float(200.0)
            type_name = search_convert_t.group(2)
            value = f'{type_name}({search_convert_t.group(3)})'
        else:
            value = res.group(2)
        name = res.group(1)
        return '{} = {}'.format(name, value)  # 'PI = 3.14159265358979323846'
    elif res := search(regex_func_def, contents):
        name = res.group(1)
        args = res.group(2)
        body = res.group(3)

        search_convert_t = re.search(r'(convert_)([\w]+)', name)
        # todo: this type of definition is not used since unparse function replaces convert_t already
        # alternative: delete define statement instead
        if search_convert_t:
            type_name = search_convert_t.group(2)
            body = '{}({})'.format(type_name, args)
        return '{} = lambda {}: {}'.format(name, args, body)  # 'MUL = lambda x, y: x * y'
    elif search(regex_include, contents):
        # until now no case occured where python needs to include something like opencl
        return ''
    elif res := search(regex_array_definition, contents):
        name, dtype, data = res.group(1), res.group(2), res.group(3)
        return f'{name} = np.array([{data}], dtype={dtype})'
    else:
        raise ValueError('Preprocessing line content not covered with Regex')


def unparse_type_def_node(node: Typedef):
    type_name_c = node.type.type.names[0]
    # return '{} = np.dtype(ClTypes.{}).type'.format(node.name, type_name_c)
    return f'{node.name} = {type_name_c}'


def search_for_barrier(code_c, ast):
    """
    Searches for barriers inside of functions.
    e.g. consider following c code (Pseudocode):
    int func_nested(){
        barrier(...);

    int func_parent(){
        func_nested()

    __kernel kernel_func(){
        func_parent

    After conversion, the barrier call is emulated by making use of yield and yield from.

    def func_nested():
        barrier(...);

    int func_parent():
        return (yield from func_nested())

    __kernel kernel_func():
        (yield from func_parent)
    """
    code_c = code_c.split('\n')
    lines_with_barrier = [line for line, line_content in enumerate(code_c)
                          if 'barrier(CLK_' in line_content]

    def get_start_line(_):
        if isinstance(_, FuncDef):
            return _.coord.line
        elif isinstance(_, list):
            return _[0].coord.line

    line_endings = [get_start_line(_) for _ in ast.ext][1:] + [len(code_c)]
    funcs = [{'name': _.decl.name,
              'start_line': _.coord.line,
              'end_line': line_endings[i],
              'is_kernel': any(['kernel' in spec for spec in _.decl.funcspec])}
             for i, _ in enumerate(ast.ext) if isinstance(_, FuncDef)]
    # or funcs with nested func that has barrier
    funcs_with_barrier = [_ for _ in funcs if
                          any(_['start_line'] < line < _['end_line'] for line in lines_with_barrier)]
    funcs_nested_barrier = [_ for _ in funcs if any(re.search(r'(\W)(' + func['name'] + r')(\W)',
                                                              ''.join(code_c[_['start_line']:_['end_line'] - 1]))
                                                    for func in funcs_with_barrier)]

    return [_['name'] for _ in funcs_with_barrier + funcs_nested_barrier if not _['is_kernel']]


def replace_builtin_macros(code_c):
    """
    remove macros with builtin function calls and replace usages:

    if macros are built in functions, this method performs the replacement where those macros are used.

    the motivation for this solution is that built-in function are accessed via the work item object wi, which cannot
    be called at the location of definement.

    #define gid0 get_global_id(0)
    #define gid1 get_global_id(1)
    #define gid2 get_global_id(2)

    // the regex search matches against the listed cases
    __kernel void my_knl(__global int *a)
    {
        a[gid0]+=1.0;
         get_gid0_
         gid0
         gid0*
         gid0/
         =gid0)
         gid0==
        ==gid0
    }
    """
    rgx = '#define[ ]+([\w]+)[ ]+(get_[\w\.\-e\(\)]+)[\s\n]'
    defines_to_be_replaced = re.findall(rgx, code_c)
    code_c = re.sub(rgx, '', code_c)
    for name, val in defines_to_be_replaced:
        code_c = re.sub('([\[\(\s\+\*\/\-\=]|==)' + f'({name})' + '([\]\)\s\+\*\/\-\;]|==)', r'\1' + val + r'\3',
                        code_c)
    return code_c


def unparse_c_code_to_python(code_c: str) -> str:
    # todo prevents files: https://stackoverflow.com/questions/12644902/how-to-prevent-table-regeneration-in-ply
    # yacc.yacc(debug=False, write_tables=False)
    code_c = re.sub('#define[ ]+TP_ROOT[ ]+(cfloat|cdouble])[ ]*(\n)', '', code_c)  # removes TP_ROOT = cfloat

    code_c = replace_builtin_macros(code_c)

    p = OpenCLCParser(lex_optimize=False, yacc_optimize=False)
    os.remove('yacctab.py')
    # remove block comments like /* some comment */ since other p.parse throws parsing error
    code_c = re.sub(r'\/\*(\*(?!\/)|[^*])*\*\/', '', code_c)
    code_c = code_c.replace('#pragma unroll', '')

    code_c = MacroWithArguments.replace_with_function(code_c)

    from pyopencl_extension.framework import preamble_activate_complex_numbers
    code_c = code_c.replace(preamble_activate_complex_numbers, '')
    from pyopencl_extension.framework import preamble_activate_double
    code_c = code_c.replace(preamble_activate_double, '')
    code_c = code_c.replace('__const', '')  # todo: create constant array class which raises error when writing to
    # todo: comments can be extracted using line numbers. Nodes in abstract syntax tree provide coords for reinsertion
    ast = p.parse(code_c)  # abstract syntax tree, why no comments? --> https://github.com/eliben/pycparser/issues/124
    module_py = []
    header = """
from typing import Tuple
from pyopencl_extension.emulation import cl_kernel, WorkItem, local_memory
from pyopencl_extension.types.funcs_for_emulation import *
from pyopencl_extension.types.utilities_np_cl import Types, c_to_np_type_name_catch
import numpy as np
            """
    module_py.append(header)
    if 'cfloat' in code_c:
        module_py.append(preamble_buff_t_complex64_np)
    elif 'cdouble' in code_c:
        module_py.append(preamble_buff_t_complex128_np)
    # module_py.append(preamble_cl_funcs_to_lambdas)

    # find funcs that contain barrier(CLK_LOCAL_MEM_FENCE) and therefore require yield from
    names_func_has_barrier.clear()
    names_func_has_barrier.extend(search_for_barrier(code_c, ast))

    names_func_require_work_item.clear()
    names_func_require_work_item.extend([node.decl.name for node in ast.ext if isinstance(node, FuncDef)])
    for node in ast.ext:
        if type(node) == list:
            if len(node) == 1:
                if type(node[0]) == PreprocessorLine:
                    module_py.append(unparse_preprocessor_line(node[0]))
        if type(node) == Typedef:
            module_py.append(unparse_type_def_node(node))
        if isinstance(node, FuncDef):
            module_py.append('\n')
            if node.decl.name in MacroWithArguments.names_py_macro:  # for explanation see comment below names_macro_func_def
                module_py.append(MacroWithArguments.unparse_macro_node_and_convert_to_string(node))
            else:
                module_py.append(unparse_function_node(node))

    code_py = '\n'.join(module_py)
    code_py = code_py + '\n'
    # todo: deal with complex header
    # if 'cfloat_t' in code_c:
    #     preamble_buff_t = preamble_buff_t_complex64_np
    # elif 'cdouble_t' in code_c:
    #     preamble_buff_t = preamble_buff_t_complex128_np
    # else:
    #     preamble_buff_t = preamble_buff_t_real_np
    #
    # preamble_buff_t = '{}\n\n{}'.format(preamble_buff_t, preamble_cl_funcs_to_lambdas)
    # logging.info(code_py)
    return code_py


b_use_existing_file_for_emulation = False


def use_existing_file_for_emulation(value: bool):
    # hack solution for making changes in generated python cl program directly. When rerunning, those changes are not
    # overridden
    global b_use_existing_file_for_emulation
    b_use_existing_file_for_emulation = value


def create_py_file_and_load_module(code_py: str, file: str = None):
    """
    Creates .py file from code_py and loads created module such that debugger can jump into code.
    :param code_py: Python code snippet in string format
    :param file: Either full path or path relative to working directory without including file type name
    :return:
    """
    if file is None:
        file = str('cl_py_modules/debug_file')
    path_code_py = '{}.py'.format(file)
    # hack solution for making changes in generated python cl program directly. When rerunning, those changes are not
    # overridden
    if Path(path_code_py).exists() and b_use_existing_file_for_emulation:
        pass
    else:
        write_string_to_file(code_py, path_code_py)

    import importlib.util
    spec = importlib.util.spec_from_file_location(file, str(path_code_py))
    program_python = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(program_python)
    return program_python


def compute_linear_idx(idx_tuple, dimensions) -> int:
    idx = 0
    n_dim = len(dimensions)
    for i in range(n_dim):
        offset = 1
        for j in range(i + 1, n_dim):
            offset *= int(dimensions[j])  # self.get_global_size(j) / self.get_local_size(j)
        idx += idx_tuple[i] * offset
    return idx


def compute_tuple_idx(idx_lin, dimensions):
    n_dim = len(dimensions)
    idx_tuple = [0] * n_dim
    for _i in range(n_dim):
        i = n_dim - 1 - _i
        idx_lin, idx_tuple[i] = divmod(idx_lin, dimensions[i])
    return tuple(idx_tuple)



__author__ = "P.Mohr"
__copyright__ = "10.02.2021, P.Mohr"
__version__ = "1.0"
__email__ = "philipp.mohr@tuhh.de"
__doc__ = """This module contains the FFT operation. Radix 2 and """

from dataclasses import dataclass
from typing import List

import numpy as np

from pyopencl_extension import Function, Kernel, Types, Scalar, \
    Global, Program, Private, Local, zeros_like, Array, zeros


# Definition of radix 2 - 16 functions:

def get_funcs_radix2(data_t):
    func_fft_radix_2 = Function('fft_radix_2',
                                {'v': Private(data_t)},
                                """
                           data_t v0= v[0];
                           v[0] = v0 + v[1];
                           v[1] = v0 - v[1];
                           """)
    return [], func_fft_radix_2


mul_pxpy_dict4 = {'mul_p1q2': (mul_p1q2 := 'return (data_t)(a.y,-a.x);')}


def get_funcs_radix4(data_t):
    # mul_pxqy(a) returns a*exp(-j * PI * p / q) where p=x and q=y
    funcs_mulpxpy_4 = [Function(k, {'a': Scalar(data_t)}, v, returns=data_t) for k, v in mul_pxpy_dict4.items()]
    func_fft_radix_4 = Function('fft_radix_4',
                                {'v': Private(data_t)},
                                """
            // 2x DFT2 and twiddle
            data_t v0 = v[0] + v[2];
            data_t v1 = v[0] - v[2];
            data_t v2 = v[1] + v[3];
            data_t v3 = mul_p1q2(v[1] - v[3]); // twiddle

            // 2x DFT2 and store
            v[0] = v0 + v2;
            v[1] = v1 + v3;
            v[2] = v0 - v2;
            v[3] = v1 - v3;
                                """)
    return funcs_mulpxpy_4, func_fft_radix_4


mul_pxpy_dict8 = {**{'mul_p0q2': (mul_p0q2 := 'return a;'),
                     'mul_p0q4': mul_p0q2,
                     'mul_p2q4': (mul_p2q4 := mul_p1q2),
                     'mul_p1q4': (mul_p1q4 := 'return (data_t)(SQRT_1_2)*(data_t)(a.x+a.y,-a.x+a.y);'),
                     'mul_p3q4': (mul_p3q4 := 'return (data_t)(SQRT_1_2)*(data_t)(-a.x+a.y,-a.x-a.y);')},
                  **mul_pxpy_dict4}


def get_funcs_radix8(data_t):
    funcs_mulpxpy_8 = [Function(k, {'a': Scalar(data_t)}, v, returns=data_t) for k, v in mul_pxpy_dict8.items()]
    func_fft_radix_8 = Function('fft_radix_8',
                                {'v': Private(data_t)},
                                """
            // 4x in-place DFT2
            data_t u0 = v[0];
            data_t u1 = v[1];
            data_t u2 = v[2];
            data_t u3 = v[3];
            data_t u4 = v[4];
            data_t u5 = v[5];
            data_t u6 = v[6];
            data_t u7 = v[7];

            data_t v0 = u0 + u4;
            data_t v4 = mul_p0q4(u0 - u4);
            data_t v1 = u1 + u5;
            data_t v5 = mul_p1q4(u1 - u5);
            data_t v2 = u2 + u6;
            data_t v6 = mul_p2q4(u2 - u6);
            data_t v3 = u3 + u7;
            data_t v7 = mul_p3q4(u3 - u7);

            // 4x in-place DFT2 and twiddle
            u0 = v0 + v2;
            u2 = mul_p0q2(v0 - v2);
            u1 = v1 + v3;
            u3 = mul_p1q2(v1 - v3);
            u4 = v4 + v6;
            u6 = mul_p0q2(v4 - v6);
            u5 = v5 + v7;
            u7 = mul_p1q2(v5 - v7);

            // 4x DFT2 and store (reverse binary permutation)
            v[0]   = u0 + u1;
            v[1]   = u4 + u5;
            v[2] = u2 + u3;
            v[3] = u6 + u7;
            v[4] = u0 - u1;
            v[5] = u4 - u5;
            v[6] = u2 - u3;
            v[7] = u6 - u7;
                                """)
    return funcs_mulpxpy_8, func_fft_radix_8


mul_pxpy_dict16 = {**{'mul_p0q8 ': mul_p0q2,
                      'mul_p1q8': 'return mul_1((data_t)(COS_8,-SIN_8),a);',
                      'mul_p2q8': mul_p1q4,
                      'mul_p3q8': 'return mul_1((data_t)(SIN_8,-COS_8),a);',
                      'mul_p4q8 ': mul_p2q4,
                      'mul_p5q8': 'return mul_1((data_t)(-SIN_8,-COS_8),a);',
                      'mul_p6q8': mul_p3q4,
                      'mul_p7q8': 'return mul_1((data_t)(-COS_8,-SIN_8),a);'}, **mul_pxpy_dict8}


def get_funcs_radix16(data_t):
    func_mul1 = Function('mul_1',
                         {'a': Scalar(data_t), 'b': Scalar(data_t)},
                         'data_t x; x.even = MUL_RE(a,b); x.odd = MUL_IM(a,b); return x;', returns=data_t,
                         defines={'MUL_RE(a,b)': '(a.even*b.even - a.odd*b.odd)',
                                  'MUL_IM(a,b)': '(a.even*b.odd + a.odd*b.even)'})

    funcs_mulpxpy_16 = [Function(k, {'a': Scalar(data_t)}, v, returns=data_t) for k, v in mul_pxpy_dict16.items()]
    funcs_mulpxpy_16[0].defines = {'COS_8': np.cos(np.pi / 8), 'SIN_8': np.sin(np.pi / 8)}
    func_fft_radix_16 = Function('fft_radix_16',
                                 {'v': Private(data_t)},
                                 """
            data_t u[16];
            for (int m=0;m<16;m++) u[m] = v[m];
            // 8x in-place DFT2 and twiddle (1)
            DFT2_TWIDDLE(u[0],u[8],mul_p0q8);
            DFT2_TWIDDLE(u[1],u[9],mul_p1q8);
            DFT2_TWIDDLE(u[2],u[10],mul_p2q8);
            DFT2_TWIDDLE(u[3],u[11],mul_p3q8);
            DFT2_TWIDDLE(u[4],u[12],mul_p4q8);
            DFT2_TWIDDLE(u[5],u[13],mul_p5q8);
            DFT2_TWIDDLE(u[6],u[14],mul_p6q8);
            DFT2_TWIDDLE(u[7],u[15],mul_p7q8);

            // 8x in-place DFT2 and twiddle (2)
            DFT2_TWIDDLE(u[0],u[4],mul_p0q4);
            DFT2_TWIDDLE(u[1],u[5],mul_p1q4);
            DFT2_TWIDDLE(u[2],u[6],mul_p2q4);
            DFT2_TWIDDLE(u[3],u[7],mul_p3q4);
            DFT2_TWIDDLE(u[8],u[12],mul_p0q4);
            DFT2_TWIDDLE(u[9],u[13],mul_p1q4);
            DFT2_TWIDDLE(u[10],u[14],mul_p2q4);
            DFT2_TWIDDLE(u[11],u[15],mul_p3q4);

            // 8x in-place DFT2 and twiddle (3)
            DFT2_TWIDDLE(u[0],u[2],mul_p0q2);
            DFT2_TWIDDLE(u[1],u[3],mul_p1q2);
            DFT2_TWIDDLE(u[4],u[6],mul_p0q2);
            DFT2_TWIDDLE(u[5],u[7],mul_p1q2);
            DFT2_TWIDDLE(u[8],u[10],mul_p0q2);
            DFT2_TWIDDLE(u[9],u[11],mul_p1q2);
            DFT2_TWIDDLE(u[12],u[14],mul_p0q2);
            DFT2_TWIDDLE(u[13],u[15],mul_p1q2);

            // 8x DFT2 and store (reverse binary permutation)
            v[0]  = u[0]  + u[1];
            v[1]  = u[8]  + u[9];
            v[2]  = u[4]  + u[5];
            v[3]  = u[12] + u[13];
            v[4]  = u[2]  + u[3];
            v[5]  = u[10] + u[11];
            v[6]  = u[6]  + u[7];
            v[7]  = u[14] + u[15];
            v[8]  = u[0]  - u[1];
            v[9]  = u[8]  - u[9];
            v[10] = u[4]  - u[5];
            v[11] = u[12] - u[13];
            v[12] = u[2]  - u[3];
            v[13] = u[10] - u[11];
            v[14] = u[6]  - u[7];
            v[15] = u[14] - u[15];
                                """,
                                 defines={'DFT2_TWIDDLE(a,b,t)': '{ data_t tmp = t(a-b); a += b; b = tmp; }'})
    funcs_helpers = [func_mul1] + funcs_mulpxpy_16
    return funcs_helpers, func_fft_radix_16


dict_radix_funcs = {
    2: get_funcs_radix2,
    4: get_funcs_radix4,
    8: get_funcs_radix8,
    16: get_funcs_radix16,
}


def get_funcs_radixes(radixes, data_t: np.dtype):
    # makes sure radixes are in order, since higher radix functions use function loaded for lower radixes.
    radixes = sorted(radixes)
    defines_radices = {'SQRT_1_2': np.cos(np.pi / 4)}
    funcs = [dict_radix_funcs[r](data_t) for r in radixes]
    # filter out duplicate helpers
    helpers = list({helper.name: helper for helpers, _ in funcs for helper in helpers}.values())
    func_radices = [func for _, func in funcs]
    return helpers + func_radices, defines_radices


def test_get_funcs_radixes():
    res = get_funcs_radixes([8, 16], Types.float2)
    assert res[0][-2].name == 'fft_radix_8'
    assert res[0][-1].name == 'fft_radix_16'


@dataclass
class FftStageBuilder:
    radix: int
    data_t: np.dtype
    real_t: np.dtype
    fft_size: int
    global_mem_cacheline_size: int
    size_data_in_first_iteration: int
    iteration_max: int
    b_local_memory: bool = False

    def get_funcs_for_all_stages(self, radixes: List[int]):
        conf = self
        typedefs = {'data_t': (data_t := conf.data_t),
                    'real_t': (real_t := conf.real_t)}

        R = conf.radix
        defines = {'R': R,
                   'N': (N := conf.fft_size),
                   # work group size:
                   'T': (T := min(int(N / R), conf.global_mem_cacheline_size)),
                   'CW': (CW := conf.global_mem_cacheline_size),  # Coalescing width,
                   'N_INPUT': conf.size_data_in_first_iteration,  # in_buffer length which might not be power of two
                   'ITERATION_MAX': self.iteration_max,  # in_buffer length which might not be power of two
                   }

        funcs_radix, defines_radices = get_funcs_radixes(radixes, data_t)
        defines = {**defines, **defines_radices}

        # [1]: The expand() function can be thought of as inserting a dimension of length N2 after the first
        # dimension of length N1 in a linearized index.
        func_expand = Function('expand',
                               {'idxL': Scalar(Types.int),
                                'N1': Scalar(Types.int),
                                'N2': Scalar(Types.int)},
                               """
                                 return (int)(idxL/N1)*N1*N2 + (idxL%N1);
                                 """,
                               returns=Types.int)
        # float2* v, int R, int idxD, int incD, int idxS, int incS
        func_exchange = Function('exchange',
                                 {'v': Private(data_t),
                                  'idxD': Scalar(Types.int),
                                  'incD': Scalar(Types.int),
                                  'idxS': Scalar(Types.int),
                                  'incS': Scalar(Types.int),
                                  'shared': Local(real_t),
                                  },
                                 """
                                __local real_t* sr = shared; // vector with real part
                                __local real_t* si = shared+T*R; // vector with imaginary part
                                barrier(CLK_LOCAL_MEM_FENCE);
                                for( int r=0; r<R; r++ ) {
                                    int i = (idxD + r*incD)*STRIDE;
                                    sr[i] = v[r].s0;
                                    si[i] = v[r].s1;
                                }
                                barrier(CLK_LOCAL_MEM_FENCE);
                                for(int r=0; r<R; r++ ) {
                                    int i = (idxS + r*incS)*STRIDE;
                                    v[r] = (data_t)(sr[i], si[i]);
                                }
                                 """,
                                 defines={'STRIDE': 1})
        func_v_from_data0 = Function('v_from_data0',
                                     {'v': Private(self.data_t),
                                      'data0': Local(self.data_t) if self.b_local_memory else Global(self.data_t,
                                                                                                     read_only=True),
                                      'idxS': Scalar(Types.int),
                                      'r': Scalar(Types.int),
                                      'direction': Scalar(Types.int),
                                      'iteration': Scalar(Types.int)},
                                     """
                 if(iteration ==0){
                     if(idxS<N_INPUT){
                        v[r]=data0[(int)(idxS)];
                     }else{
                        v[r]=(data_t)(0.0, 0.0);
                     }
                     if(direction == -1){
                        v[r] = (data_t)(v[r].s1, v[r].s0);
                     }
                 }else{
                    v[r] = data0[(int)(idxS)];
                 }
                 """)
        func_data_1_from_v = Function('data_1_from_v',
                                      {'v': Private(self.data_t),
                                       'data1': Local(self.data_t) if self.b_local_memory else Global(self.data_t),
                                       'idxD': Scalar(Types.int),
                                       'r': Scalar(Types.int),
                                       'direction': Scalar(Types.int),
                                       'iteration': Scalar(Types.int)},
                                      """
                 if(iteration==ITERATION_MAX && direction == -1){
                    data1[idxD] = (data_t)((float)(1.0/N)) * (data_t)(v[r].s1, v[r].s0);
                 }else{
                    data1[idxD] = v[r];
                 }
                 """)
        shared_memory_init = 'local real_t shared[(int)(2*T*R)];//2*Ns];//T*R];'
        return funcs_radix + [func_exchange, func_expand, func_v_from_data0, func_data_1_from_v], \
               typedefs, defines, shared_memory_init

    def get_fft_stage_func(self):
        replacements = {'fft_radix_R': f'fft_radix_{self.radix}'}
        func_fft_iteration = Function('fft_stage', {
            # thread index: T radix R fft ops are done in parallel
            't': Scalar(Types.int),
            # thread block index: identifies block of N/(R*T) blocks.
            'b': Scalar(Types.int),
            'Ns': Scalar(Types.int),
            'data0': Local(self.data_t) if self.b_local_memory else Global(self.data_t, read_only=True),
            'data1': Local(self.data_t) if self.b_local_memory else Global(self.data_t),
            'shared': Local(self.real_t),
            'direction': Scalar(Types.int),
            'iteration': Scalar(Types.int)},
                                      """
        int j = b*T + t; // as proposed in [1, p.3, text section A]
        private data_t v[R];        
        real_t angle = -2*PI*(j % Ns)/(Ns*R);
        for(int r=0; r<R; r++){
            int idxS = j + r* N/R; // idxS=idxSource
            v_from_data0(v, data0, idxS, r, direction,iteration);
            real_t cos_angle = cos(r*angle);
            real_t sin_angle = sin(r*angle);
            v[r] = (data_t)(v[r].s0*cos_angle-v[r].s1 *sin_angle,
                            v[r].s0*sin_angle+v[r].s1 *cos_angle);
        }
        ${fft_radix_R}(v);

        if(Ns>=CW){ // todo: remove condition and store as separate kernel
            // changed line 27 of [1, Figure 2] to work with global dimensions of this class:
            int offset = expand(j, Ns, R); 
            for(int r=0; r<R; r++){
                int idxD = offset + r*Ns; // idxD=idxDestination
                data_1_from_v(v, data1, idxD, r, direction, iteration);
            }
        }else{ // According to [1, p.4], such that global memory writes are coalesced
            // !! mistake in [1] where *Ns is missing: idxD = (int)(t/Ns)*R + (t%Ns); !!
            int idxD = (int)(t/Ns)*Ns*R + (t%Ns);
            exchange( v, idxD, Ns, t, T, shared);
            int offset = b*R*T+ t; 
            for( int r=0; r<R; r++ ){
                idxD = offset + r*T;
                data_1_from_v(v, data1, idxD, r, direction, iteration);
            }  
        }
                   """,
                                      replacements=replacements)
        return func_fft_iteration


class FftBase:
    """
    Implementation of reference algorithm [1, Fig. 2] (where only radix 2 kernel is provided)
    A reference for higher radix kernels can be found in [2]

    [1] N. K. Govindaraju, B. Lloyd, Y. Dotsenko, B. Smith, und J. Manferdelli, „High performance discrete Fourier
    transforms on graphics processors“, in 2008 SC - International Conference for High Performance Computing,
    Networking, Storage and Analysis, Austin, TX, Nov. 2008, S. 1–12, doi: 10.1109/SC.2008.5213922.

    [2] http://www.bealto.com/gpu-fft_opencl-2.html
    """

    def __call__(self, *args, **kwargs):
        data_in = self._in_buffer
        data_out = self._data_out
        Ns = 1
        for iteration, radix in enumerate(self.schedule):
            self.kernels_for_iteration[iteration](Ns=Ns,
                                                  data_in=data_in,
                                                  data_out=data_out,
                                                  direction=self.direction,
                                                  iteration=iteration)
            Ns = Ns * radix
            if iteration == 0:
                data_in = self._data_in  # map data_in to internal buffer with correct power of two size
            data_in, data_out = data_out, data_in  # swap
        return data_in.view(self.in_buffer.dtype).reshape(self._out_shape)

    @staticmethod
    def get_mixed_radix_schedule(N, radixes):
        _radixes = [r for r in radixes if r <= int(N / 2)]  # if N is very short do not use large radix ffts
        schedule = []
        n = N
        while n != 1:
            _radix = radixes[np.argmax([(n / i).is_integer() for i in radixes])]
            schedule.append(_radix)
            n /= _radix
            if not n.is_integer():
                raise ValueError('Something went wrong. n must be integer')
        return schedule  # [2 for _ in range(int(log(N, 2)))]

    @staticmethod
    def _get_kernel(stage_builder: FftStageBuilder, iteration, radix, data_in, data_out, emulate=False):
        if iteration == 0:  # data0 is in_buffer, whose length might not be power of two
            offset_data_in = 'get_global_id(0)*N_INPUT'
        else:
            offset_data_in = 'get_global_id(0)*N'
        stage_builder.radix = radix
        funcs, type_defs, d, shared_mem_fft_stage = stage_builder.get_funcs_for_all_stages([radix])
        funcs.append(stage_builder.get_fft_stage_func())
        d['M'] = (M := data_in.shape[0])  # number of FFTs to perform simultaneously
        knl_gpu_fft = Kernel('gpu_fft',
                             {'Ns': Scalar(Types.int),
                              # In each iteration, the algorithm can be thought of combining the radix R FFTs on
                              # subsequences of length Ns into the FFT of a new sequence of length RNs by
                              # performing an FFT of length R on the corresponding elements of the subsequences.
                              'data_in': Global(data_in),
                              'data_out': Global(data_out),
                              'direction': Scalar(Types.int),
                              'iteration': Scalar(Types.int)},
                             """
                 ${shared_mem_fft_stage}
                 int t = get_local_id(2); // thread index 
                 int b = get_global_id(1); // thread block index
                 fft_stage(t, b, Ns, 
                           data_in  +${offset_data_in},
                           data_out +get_global_id(0)*N,
                           shared,direction, iteration);""",
                             replacements={'offset_data_in': offset_data_in,
                                           'shared_mem_fft_stage': shared_mem_fft_stage},
                             global_size=(d['M'], max(1, int(d['N'] / (d['R'] * d['T']))), d['T']),
                             local_size=(1, 1, d['T']))
        program = Program(funcs,
                          [knl_gpu_fft], defines=d, type_defs=type_defs
                          ).compile(context=data_in.context, emulate=emulate,
                                    file=Program.get_default_dir_pycl_kernels().joinpath(
                                        f'fft_{iteration}_{stage_builder.radix}'))
        return program.gpu_fft

    def __init__(self, in_buffer: Array, b_inverse_fft=False, emulate=False):
        # just for debugging:
        self.emulate = emulate
        self.direction = -1 if b_inverse_fft else +1

        # fir internal data types to input data type
        if in_buffer.dtype == Types.cfloat:
            data_t = Types.float2
            real_t = Types.double
        elif in_buffer.dtype == Types.cdouble:
            data_t = Types.double2
            real_t = Types.double
        else:
            raise ValueError()

        # input ist zero padded such that its length becomes a power of 2.
        # There exists fast FFT versions for non power of two, but those are currently not implemented.
        # For further information read here:
        # https://math.stackexchange.com/questions/77118/non-power-of-2-ffts
        # https://stackoverflow.com/questions/13841296/fft-in-numpy-python-when-n-is-not-a-power-of-2
        if in_buffer.ndim == 1:
            b_1d_input = True
            in_buffer = in_buffer.reshape((1, in_buffer.shape[0]))
        else:
            b_1d_input = False
        N = in_buffer.shape[1]
        M = in_buffer.shape[0]
        if not np.log2(N).is_integer():  # if not power of 2, pad accordingly
            N = 2 ** int(np.log2(N) + 1)
        if b_1d_input:
            self._out_shape = (N,)
        else:
            self._out_shape = (M, N)

        data_in = zeros((M, N), data_t)
        data_out = zeros_like(data_in)

        radixes = [16, 8, 4, 2]  # available radix ffts
        self.schedule = self.get_mixed_radix_schedule(N, radixes)
        fft_stage_builder = FftStageBuilder(
            radix=2,  # might change according to schedule
            data_t=data_t, real_t=real_t,
            fft_size=N,
            global_mem_cacheline_size=in_buffer.context.devices[0].global_mem_cacheline_size,
            size_data_in_first_iteration=in_buffer.shape[1], iteration_max=len(self.schedule) - 1)
        self.kernels_for_iteration = [
            self._get_kernel(fft_stage_builder, iteration=_iter, radix=_radix, data_in=data_in, data_out=data_out,
                             emulate=emulate) for _iter, _radix
            in enumerate(self.schedule)]
        self.in_buffer = in_buffer
        self._in_buffer = in_buffer.view(data_t)
        self._data_in = data_in
        self._data_out = data_out

        # optimize later: there may be some redundant kernels in self.kernels_for_iteration
        # max_iteration = len(self.schedule) - 1
        # # individual kernels for: iteration 0 and each different radix in other iterations
        # iter_unique_knl = [0] + (1 + np.sort(np.unique(self.schedule[1:], return_index=True)[1])).tolist() + \
        #                   [max_iteration + 1]
        # _parts = [[_iter, self.schedule[_iter]] for _iter in iter_unique_knl[:-1]]
        # _knl_ids = np.concatenate([np.ones(iter_unique_knl[i + 1] - iter_unique_knl[i], int) * i
        #                            for i, _ in enumerate(iter_unique_knl[:-1])])
        # if b_inverse_fft:  # final iteration has individual kernel
        #     pass


class Fft(FftBase):
    def __init__(self, in_buffer: Array, emulate=False):
        super().__init__(in_buffer, emulate=emulate)


class IFft(FftBase):
    def __init__(self, in_buffer: Array, emulate=False):
        super().__init__(in_buffer, b_inverse_fft=True, emulate=emulate)



import numpy as np

from pyopencl_extension import SumAlongAxis, to_device


def test_sum_along_axis():
    ary = np.array([[1, 2, 3], [1, 2, 3]])
    ary_buffer = to_device(ary)
    """"
    sum_along_axis = SumAlongAxis(ary_buffer, axis=0)
    res = sum_along_axis().get()
    ref = ary.sum(axis=0)
    assert np.all(res == ref)
    """

    sum_along_axis = SumAlongAxis(ary_buffer, axis=1)
    res = sum_along_axis().get()
    ref = ary.sum(axis=1)
    assert np.all(res == ref)


def test_sum_along_axis_1d():
    ary = np.array([1, 2, 3])
    ary_buffer = to_device(ary)

    sum_along_axis = SumAlongAxis(ary_buffer, axis=0)
    # res_py = sum_along_axis(emulate=True).get()
    res_cl = sum_along_axis().get()
    ref = ary.sum(axis=0)
    assert np.all(res_cl == ref)



import time

import pyopencl as cl
import pyopencl.array as cl_array
from pyopencl.array import Array as ArrayCl # avoids warning in PyCharm IDE for missing attrbutes
from pyopencl_extension.modifications_pyopencl.command_queue import CommandQueue, get_current_queue


class Array(ArrayCl):
    @classmethod
    def from_array(cls, array: ArrayCl):
        if isinstance(array, ArrayCl):
            return Array((a := array).queue, a.shape, a.dtype, order="C", allocator=a.allocator,
                         data=a.data, offset=a.offset, strides=a.strides, events=a.events)
        elif isinstance(array, Array):
            return array

    def set(self, ary, queue=None, async_=None, **kwargs):
        res = super().set(ary, queue, async_, **kwargs)
        self.add_latest_event('set')
        return res

    def get(self, queue=None, ary=None, async_=None, **kwargs):
        t0 = time.perf_counter_ns()
        res = super().get(queue, ary, async_, **kwargs)
        self.queue.t_ns.add_blocking(time.perf_counter_ns() - t0)
        self.add_latest_event('get')
        return res

    def add_latest_event(self, name):
        if len(self.events) > 0:
            # [self.queue.add_event(event, name) for event in self.events]
            self.queue.add_event(self.events[-1], name)

    def view(self, dtype=None):
        return self.from_array(super().view(dtype))


def to_device(ary, queue=None, allocator=None, async_=None, array_queue=cl_array._same_as_transfer, **kwargs):
    queue = get_current_queue() if queue is None else queue
    res = Array.from_array(cl_array.to_device(queue, ary, allocator, async_, array_queue, **kwargs))
    if len(res.events) > 0:
        queue.add_event(res.events[-1], 'to_device')
    return res


def empty(shape, dtype, cq: CommandQueue = None, order="C", allocator=None,
          data=None, offset=0, strides=None, events=None, _flags=None):
    cq = get_current_queue() if cq is None else cq
    res = Array.from_array(ArrayCl(cq, shape, dtype, order, allocator))
    res.add_latest_event('empty')
    return res


def zeros(shape, dtype, queue: CommandQueue = None, order="C", allocator=None):
    queue = get_current_queue() if queue is None else queue
    res = Array.from_array(cl_array.zeros(queue, shape, dtype, order, allocator))
    res.add_latest_event('zeros')
    return res


def empty_like(ary, queue=cl_array._copy_queue, allocator=None):
    res = Array.from_array(cl_array.empty_like(ary, queue, allocator))
    res.add_latest_event('empty_like')
    return res


def zeros_like(ary):
    res = Array.from_array(cl_array.zeros_like(ary))
    res.add_latest_event('zeros_like')
    return res



from pathlib import Path

whole_signed_number_types = ['char', 'short', 'int', 'long']
whole_unsigned_number_types = [f'u{t}' for t in whole_signed_number_types]
whole_number_types = whole_signed_number_types + whole_unsigned_number_types
fractional_number_types = ['half', 'float', 'double']
all_scalar_number_types = whole_number_types + fractional_number_types
vector_type_lengths = [2, 4, 8, 16]
all_vec_number_types = [f'{num}{vec}' for num in all_scalar_number_types
                        for vec in vector_type_lengths]
all_number_types = all_scalar_number_types + all_vec_number_types
auto_gen_dir = Path(__file__).parent


def setup_file_cl_types():
    # preamble_typedefs_np_c = '\n'.join([f'{v} = np.dtype(\'{k}\').type' for k, v in np_to_c_type_name.items()])
    fields = '\n'.join([f'   {t}:Union[np.dtype, Callable]=cltypes.{t}' for t in all_vec_number_types])
    class_def1 = '@dataclass(frozen=True)\n' \
                 'class ClTypesVector:\n' + fields + '\n'

    fields = '\n'.join([f'   {t}:Union[np.dtype, Callable]=cltypes.{t}' for t in all_scalar_number_types])
    class_def2 = '@dataclass(frozen=True)\n' \
                 'class ClTypesScalar:\n' + fields + '\n'

    class_def3 = '@dataclass(frozen=True)\n' \
                 'class _ClTypes(ClTypesScalar, ClTypesVector):\n' \
                 '   pass' + '\n'

    class_defs = '\n'.join([class_def1, class_def2, class_def3])

    import_cltypes = 'from pyopencl_extension.modifications_pyopencl import cltypes\n' \
                     'from dataclasses import dataclass\n' \
                     'import numpy as np\n' \
                     'from typing import Callable, Union\n'
    content = '\n'.join([import_cltypes, class_defs])
    with open(auto_gen_dir.joinpath('cl_types.py'), 'w+') as file:
        file.write(content)


def setup_file_cl_types_for_import():
    # preamble_typedefs_np_c = '\n'.join([f'{v} = np.dtype(\'{k}\').type' for k, v in np_to_c_type_name.items()])
    fields = '\n'.join([f'   {t}:Union[np.dtype, Callable]=cltypes.{t}' for t in all_vec_number_types])
    class_def1 = '@dataclass(frozen=True)\n' \
                 'class ClTypesVector:\n' + fields + '\n'

    vector = '\n'.join([f'{t}=cltypes.{t}' for t in all_vec_number_types])
    scalar = '\n'.join([f'{t}=cltypes.{t}' for t in all_scalar_number_types])

    import_cltypes = 'from pyopencl_extension.modifications_pyopencl import cltypes\n' \
                     'import numpy as np'

    content = '\n'.join([import_cltypes, vector , scalar,
                         """
cfloat=np.dtype(np.complex64)
cdouble=np.dtype(np.complex128)
                         """])
    with open(auto_gen_dir.joinpath('cl_types_import.py'), 'w+') as file:
        file.write(content)


def setup_file_include_for_emulation():
    # preamble_typedefs_np_c = '\n'.join([f'{v} = np.dtype(\'{k}\').type' for k, v in np_to_c_type_name.items()])
    python_types = '\n'.join(['int_ = int', 'float_=float'])
    scalar = '\n'.join(
        [f'{t}=TypeHandlerScalar(cltypes.{t})' for t in all_scalar_number_types])
    vec = '\n'.join([f'{t}=TypeHandlerVec(\'{t}\')' for t in all_vec_number_types])
    import_cltypes = 'from pyopencl_extension.modifications_pyopencl import cltypes\n' \
                     'from pyopencl_extension.types.type_handler import *'
    content = '\n'.join([import_cltypes, python_types, scalar, vec])
    with open(auto_gen_dir.joinpath('types_for_emulation.py'), 'w+') as file:
        file.write(content)


if __name__ == '__main__':
    setup_file_cl_types_for_import()
    setup_file_include_for_emulation()
    setup_file_cl_types()
